# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class RadioStationEntry:

    def __init__(self, id : int, codec : str, url : str, title : str, mimetype : str, bitrate : int = None):
        self._id : int = id
        self._title : str = title
        self._url : str = url
        self._codec : str = codec
        self._mimetype : str = mimetype
        self._bitrate : int = bitrate

    @property
    def id(self) -> int:
        return self._id

    @property
    def title(self) -> str:
        return self._title

    @property
    def url(self) -> str:
        return self._url
    
    @property
    def codec(self) -> str:
        return self._codec

    @property
    def mimetype(self) -> str:
        return self._mimetype

    @property
    def bitrate(self) -> int:
        return self._bitrate



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cmdtalkplugin
import datetime
import config
import constants

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()


class SimpleMsgProcessor(cmdtalkplugin.Processor):

    def __init__(self, dispatcher):
        super().__init__(dispatcher)
        self.__append_timestamp: bool = config.get_config_param_as_bool(constants.ConfigParam.LOG_WITH_TIMESTAMP)

    def log(self, s):
        if self.__append_timestamp:
            super().log(f"{datetime.datetime.now()} {s}")
        else:
            super().log(s)


msgproc: SimpleMsgProcessor = SimpleMsgProcessor(dispatcher=dispatcher)



#!/usr/bin/env python3

# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os

from typing import Callable

import html
import cmdtalkplugin
import upmplgutils

import codec
import identifier_util
import radio_stations
import upnp_util

from item_identifier import ItemIdentifier
from item_identifier_key import ItemIdentifierKey
from tag_type import TagType, get_tag_type_by_name
from element_type import ElementType

from radio_station_entry import RadioStationEntry

import constants

# Prefix for object Ids. This must be consistent with what contentdirectory.cxx does
_g_myprefix = f"0${constants.plugin_name}$"
upmplgutils.setidprefix(constants.plugin_name)

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)

# Possible once initialisation. Always called by browse() or search(), should remember if it has
# something to do (e.g. the _g_init thing, but this could be something else).
_g_init = False
def _init_mother_earth_radio():
    global _g_init
    if _g_init:
        return True

    # Do whatever is needed here
    msgproc.log(f"Mother Earth Radio Plugin Release {constants.plugin_version}")

    _g_init = True
    return True

def build_intermediate_url(rp_id : int) -> str:
    http_host_port = os.environ["UPMPD_HTTPHOSTPORT"]
    url = f"http://{http_host_port}/{constants.plugin_name}/track/version/1/trackId/{str(rp_id)}"
    msgproc.log(f"intermediate_url for rp_id {rp_id} -> [{url}]")
    return url

def build_streaming_url(station_id : str) -> str:
    current : RadioStationEntry
    for current in radio_stations.radio_station_list:
        msgproc.log(f"build_streaming_url current.id: [{current.id}] station_id [{station_id}] [{type(station_id)}]")
        if str(current.id) == station_id:
            msgproc.log(f"build_streaming_url for station_id: [{station_id}] -> [{current.url}]")
            return current.url
    msgproc.log(f"build_streaming_url not found for station_id: [{station_id}]")
    return None

@dispatcher.record('trackuri')
def trackuri(a):
    upmpd_pathprefix = os.environ["UPMPD_PATHPREFIX"]
    msgproc.log(f"UPMPD_PATHPREFIX: [{upmpd_pathprefix}] trackuri: [{a}]")
    rp_id = upmplgutils.trackid_from_urlpath(upmpd_pathprefix, a)
    url = build_streaming_url(rp_id) or ""
    msgproc.log(f"intermediate_url for rp_id {rp_id} -> [{url}]")
    return {'media_url' : url}

def _returnentries(entries):
    """Helper function: build plugin browse or search return value from items list"""
    return {"entries" : json.dumps(entries), "nocache" : "0"}

def _objidtopath(objid):
    if objid.find(_g_myprefix) != 0:
        raise Exception(f"subsonic: bad objid {objid}: bad prefix")
    return objid[len(_g_myprefix):].lstrip("/")

def radio_entry_data_to_entry(objid, entry_id : str, radio_station_entry : RadioStationEntry) -> dict:
    entry : dict = {}
    entry['id'] = entry_id
    entry['pid'] = radio_station_entry.id
    entry['tp']= 'it'
    entry['uri'] = build_intermediate_url(radio_station_entry.id)
    upnp_util.set_class('object.item.audioItem.audioBroadcast', entry)
    title : str = f"{radio_station_entry.title} [{radio_station_entry.codec}]"
    upnp_util.set_album_title(title, entry)
    upnp_util.set_artist(radio_station_entry.title, entry)
    entry['upnp:album'] = radio_station_entry.codec
    entry['res:mime'] = radio_station_entry.mimetype
    upnp_util.set_bit_depth(radio_station_entry.bit_depth, entry)
    upnp_util.set_sample_rate(radio_station_entry.sampling_rate, entry)
    upnp_util.set_channel_count(radio_station_entry.channel_count, entry)
    upnp_util.set_bit_rate(
        upnp_util.calc_bitrate(
            radio_station_entry.channel_count,
            radio_station_entry.bit_depth,
            radio_station_entry.sampling_rate),
        entry)
    return entry

def radio_station_to_entry(objid, radio_station_entry : RadioStationEntry) -> dict:
    identifier : ItemIdentifier = ItemIdentifier(ElementType.RADIO_ENTRY.getName(), str(radio_station_entry.id))
    id : str = identifier_util.create_objid(
        objid = objid, 
        id = identifier_util.create_id_from_identifier(identifier))
    return radio_entry_data_to_entry(objid = objid, entry_id = id, radio_station_entry = radio_station_entry)

def get_radio_by_id(radio_id : str) -> RadioStationEntry:
    current : RadioStationEntry
    for current in radio_stations.radio_station_list:
        if radio_id == str(current.id): return current
    return None

def handler_tag_by_something(
        objid, 
        thing_extractor : Callable[[RadioStationEntry], str],
        element_to_be_created : ElementType,
        entries : list) -> list:
    thing_set : set[str] = set()
    current : RadioStationEntry
    for current in radio_stations.radio_station_list:
        curr_thing : str = thing_extractor(current)
        if not curr_thing in thing_set:
            # add entry
            identifier : ItemIdentifier = ItemIdentifier(element_to_be_created.getName(), curr_thing)
            id : str = identifier_util.create_objid(objid, identifier_util.create_id_from_identifier(identifier))
            entry : dict = upmplgutils.direntry(
                id = id, 
                pid = objid, 
                title = curr_thing)
            entries.append(entry)
            # store in set
            thing_set.add(curr_thing)
    return entries

def handler_tag_by_codec(
        objid, 
        item_identifier : ItemIdentifier, 
        entries : list) -> list:
    thing_extractor : Callable[[RadioStationEntry], str] = lambda x : x.codec
    return handler_tag_by_something(
        objid = objid,
        thing_extractor = thing_extractor,
        element_to_be_created = ElementType.ENTRY_BY_CODEC,
        entries = entries)

def handler_tag_by_name(
        objid, 
        item_identifier : ItemIdentifier, 
        entries : list) -> list:
    thing_extractor : Callable[[RadioStationEntry], str] = lambda x : x.title
    return handler_tag_by_something(
        objid = objid,
        thing_extractor = thing_extractor,
        element_to_be_created = ElementType.ENTRY_BY_TITLE,
        entries = entries)

def handler_tag_all_streams(
        objid, 
        item_identifier : ItemIdentifier, 
        entries : list) -> list:
    current : RadioStationEntry
    for current in radio_stations.radio_station_list:
        entries.append(radio_station_to_entry(objid, current))
    return entries

def handler_element_by_something(
        objid, 
        selected : str, 
        thing_extractor : Callable[[RadioStationEntry], str],
        entries : list) -> list:
    current : RadioStationEntry
    for current in radio_stations.radio_station_list:
        if thing_extractor(current) == selected:
            entries.append(radio_station_to_entry(objid, current))
    return entries

def handler_element_by_title(
        objid, 
        item_identifier : ItemIdentifier, 
        entries : list) -> list:
    select_thing : str = item_identifier.get(ItemIdentifierKey.THING_VALUE)
    thing_extractor : Callable[[RadioStationEntry], str] = lambda x : x.title
    return handler_element_by_something(objid = objid, selected = select_thing, thing_extractor = thing_extractor, entries = entries)

def handler_element_by_codec(
        objid, 
        item_identifier : ItemIdentifier, 
        entries : list) -> list:
    select_thing : str = item_identifier.get(ItemIdentifierKey.THING_VALUE)
    thing_extractor : Callable[[RadioStationEntry], str] = lambda x : x.codec
    return handler_element_by_something(objid = objid, selected = select_thing, thing_extractor = thing_extractor, entries = entries)

def handler_element_radio_entry(
        objid, 
        item_identifier : ItemIdentifier, 
        entries : list) -> list:
    radio_id : str = item_identifier.get(ItemIdentifierKey.THING_VALUE)
    msgproc.log(f"handler_element_radio_entry radio_id {radio_id}")
    radio_entry : RadioStationEntry = get_radio_by_id(radio_id)
    msgproc.log(f"handler_element_radio_entry radio_entry found [{'yes' if radio_entry else 'no'}]")
    if not radio_entry: return entries
    identifier : ItemIdentifier = ItemIdentifier(ElementType.RADIO_ENTRY.getName(), radio_id)
    id : str = identifier_util.create_objid(objid, identifier_util.create_id_from_identifier(identifier))
    track_entry : dict = radio_entry_data_to_entry(objid, id, radio_entry)
    entries.append(track_entry)
    return entries

__tag_action_dict : dict = {
    TagType.ALL_STREAMS.getTagName(): handler_tag_all_streams,
    TagType.BY_CODEC.getTagName(): handler_tag_by_codec,
    TagType.BY_TITLE.getTagName(): handler_tag_by_name,
}

__elem_action_dict : dict = {
    ElementType.ENTRY_BY_TITLE.getName(): handler_element_by_title,
    ElementType.ENTRY_BY_CODEC.getName(): handler_element_by_codec,
    ElementType.RADIO_ENTRY.getName(): handler_element_radio_entry
}

def tag_to_entry(objid, tag : TagType) -> dict[str, any]:
    tagname : str = tag.getTagName()
    identifier : ItemIdentifier = ItemIdentifier(ElementType.TAG.getName(), tagname)
    id : str = identifier_util.create_objid(
        objid = objid, 
        id = identifier_util.create_id_from_identifier(identifier))
    entry : dict = upmplgutils.direntry(
        id = id, 
        pid = objid, 
        title = get_tag_type_by_name(tag.getTagName()).getTagTitle())
    return entry

def show_tags(objid, entries : list) -> list:
    for tag in TagType:
        tag_entry : dict[str, any] = tag_to_entry(objid, tag)
        entries.append(tag_entry)
    return entries

@dispatcher.record('browse')
def browse(a):
    msgproc.log(f"browse: args: --{a}--")
    _init_mother_earth_radio()
    if 'objid' not in a: raise Exception("No objid in args")
    objid = a['objid']
    path = html.unescape(_objidtopath(objid))
    msgproc.log(f"browse: path: --{path}--")
    path_list : list[str] = objid.split("/")
    curr_path : str
    for curr_path in path_list:
        if not _g_myprefix == curr_path:
            msgproc.log(f"browse: current_path [{curr_path}] decodes to [{codec.decode(curr_path)}]")
    last_path_item : str = path_list[len(path_list) - 1] if path_list and len(path_list) > 0 else None
    msgproc.log(f"browse: path_list: --{path_list}-- last: --{last_path_item}--")
    entries = []
    if len(path_list) == 1 and _g_myprefix == last_path_item:
        # show tags
        entries = show_tags(objid, entries)
    else:
        # decode
        decoded_path : str = codec.decode(last_path_item)
        item_dict : dict[str, any] = json.loads(decoded_path)
        item_identifier : ItemIdentifier = ItemIdentifier.from_dict(item_dict)
        thing_name : str = item_identifier.get(ItemIdentifierKey.THING_NAME)
        thing_value : str = item_identifier.get(ItemIdentifierKey.THING_VALUE)
        msgproc.log(f"browse: item_identifier name: --{thing_name}-- value: --{thing_value}--")
        if ElementType.TAG.getName() == thing_name:
            tag_handler = __tag_action_dict[thing_value] if thing_value in __tag_action_dict else None
            msgproc.log(f"browse: should serve tag [{thing_value}], handler found: [{'yes' if tag_handler else 'no'}]")
            if tag_handler:
                entries = tag_handler(objid, item_identifier, entries)
                return _returnentries(entries)
        else: # it's an element
            elem_handler = __elem_action_dict[thing_name] if thing_name in __elem_action_dict else None
            msgproc.log(f"browse: should serve element [{thing_name}], handler found: [{'yes' if elem_handler else 'no'}]")
            if elem_handler:
                entries = elem_handler(objid, item_identifier, entries)
    return _returnentries(entries)


@dispatcher.record('search')
def search(a):
    msgproc.log("search: [%s]" % a)
    _init_mother_earth_radio()
    objid = a["objid"]
    entries = []

    # Run the search and build a list of entries in the expected format. See for example
    # ../radio-browser/radiotoentry for an example
    
    # msgproc.log(f"browse: returning --{entries}--")
    return _returnentries(entries)


msgproc.mainloop()



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum


class CoverSource(Enum):
    ARTIST = "artist"
    ALBUM = "album"
    SONG = "song"


def get_cover_source_by_name(name: str) -> CoverSource:
    curr: CoverSource
    for curr in CoverSource:
        if name == curr.value:
            return curr
    raise Exception(f"Name {name} could not be found")


class ArtistAlbumCoverArt:

    def __init__(
            self,
            artist_id: str,
            cover_source: CoverSource,
            object_id: str,
            cover_art: str):
        self.__artist_id: str = artist_id
        self.__cover_source: CoverSource = cover_source
        self.__object_id: str = object_id
        self.__cover_art: str = cover_art

    @property
    def artist_id(self) -> str:
        return self.__artist_id

    @property
    def cover_source(self) -> CoverSource:
        return self.__cover_source

    @property
    def object_id(self) -> str:
        return self.__object_id

    @property
    def cover_art(self) -> str:
        return self.__cover_art


class AlbumPropertyValueSelection:

    def __init__(
            self,
            album_property_key: str,
            album_property_value: str,
            album_count: int):
        self.__album_property_key: str = album_property_key
        self.__album_property_value: str = album_property_value
        self.__album_count: int = album_count

    @property
    def album_property_key(self) -> str:
        return self.__album_property_key

    @property
    def album_property_value(self) -> str:
        return self.__album_property_value

    @property
    def album_count(self) -> int:
        return self.__album_count



# Copyright (C) 2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class TidalPageDefinition(Enum):

    RISING = "rising", "pages/rising", "Tidal Rising"
    NEW_MUSIC = "newmusic", "pages/explore_new_music", "New Music"
    HOME = "home", "pages/home", "Home"
    GENRES = "genres", "pages/genre_page", "Genres"
    LOCAL_GENRES = "local_genres", "pages/genre_page_local", "Local Genres"
    MOODS = "moods", "pages/moods", "Moods"
    FOR_YOU = "foryou", "pages/for_you", "For You"
    HI_RES = "hires", "pages/hires", "Hi-Res"
    EXPLORE = "explore", "pages/explore", "Explore"

    def __init__(
            self,
            _: any,
            page_path: str,
            page_title: str):
        self.__page_path: str = page_path
        self.__page_title: str = page_title

    @property
    def page_path(self) -> str:
        return self.__page_path

    @property
    def page_title(self) -> str:
        return self.__page_title



#!/usr/bin/env python3
# Checked to run with either Python 2 or 3
#
# Copyright (C) 2017-2018 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

# Updated/improved in 2020 by Pete Callaghan.

# Handling the Flac channel from radio-paradise. This is made of
# 'blocks' which are multi-title periods with a single audio
# url. Metadata should be displayed according to the elapsed time in
# the audio (each metadata entry in the array gives the start and
# duration for the song in mS).  If metadata is re-fetched during the
# block, it is front-truncated, more or less according to the current
# time, and it may happen that the metadata for the song playing
# locally is not there any more (someone starting at this point would
# begin with the next point). The thing which does not change for
# successive fetches of the same block is the "end_event" value. We
# cache the first fetch for any block so that we are sure to have the
# right metadata available when upmpdcli calls us


from __future__ import print_function

import requests
import json
import sys
import os

PY3 = sys.version > '3'

tmpname = '/tmp/up-rp-pldata.json'

def debug(x):
    print("radio-paradise-get-flac.py: %s" % x, file=sys.stderr)
    pass

# Write new block data to cache. We always output the audio url in this case.
def newcache(jsd):
    global out
    s = json.dumps(jsd,sort_keys=True, indent=4)
    open(tmpname, 'wb').write(s.encode('utf-8'))
    os.chmod(tmpname, 0o666)
    out["audioUrl"] = jsd['url'] + "?src=alexa"

def makebytes(data):
    if type(data) == type(u''):
        return data.encode("UTF-8", errors='replace')
    return data

# Write the selected song to the output
def writesong(jsd, song):
    global out
    songs = jsd['song']
    out['title'] = songs[song]['title']
    out['artist'] = songs[song]['artist']
    if 'album' in songs[song]:
        out['album'] = songs[song]['album']
    out['artUrl'] = 'http:%s%s' % (jsd['image_base'],
                                  songs[song]['cover'])

##### Main script

    
# We're expecting args as successive pairs of "name value", and we
# need elapsedms (elapsed milliseconds in current track)
# (e.g. myscript elapsedms 134)
if (len(sys.argv)+1) % 2:
    debug("argv len not odd")
    sys.exit(1)
args = {}
for i in range(len(sys.argv)-2):
    nm = sys.argv[i+1]
    val = sys.argv[i+2]
    args[nm] = val

elapsedms = -1
try:
    elapsedms = int(args["elapsedms"])
except:
    pass

# Default value for "channel" if not set in radio.conf
# 0: main, "RP Main Mix"
# 1: mellow, "RP Mellow Mix"
# 2: rock, "RP Rock Mix"
# 3: eclectic, "RP World/Etc Mix"
if "channel" not in args:
    args["channel"] = "0"
    
debug("rp-get-flac: got elapsed %d" % elapsedms)

# Try to read the current cached data, and extract the last event (song) in the block.
cached = None
endEvent = None
try:
    s = open(tmpname, 'rb').read().decode('utf-8', errors = 'replace')
    cached = json.loads(s)

    if cached and 'end_event' in cached and elapsedms >= 0:
        endEvent = cached['end_event']

except Exception as err:
    debug("No cached data read: %s" % err)
    pass

# If we have an end event (last song), then the next block is retrieved by sending this end event
# Otherwise we just get the current block
if endEvent:
    r = requests.get("https://api.radioparadise.com/api/get_block",
                 params={"bitrate": "4", "info":"true", "chan":args["channel"], "event":endEvent})
else:
    r = requests.get("https://api.radioparadise.com/api/get_block",
                 params={"bitrate": "4", "info":"true", "chan":args["channel"]})
r.raise_for_status()
newjsd = r.json()

out = {}

# If we are currently playing, check if our cached data is still
# relevant. If it is, it is more complete than the new data which is
# front-truncated, so we use it, so that we can check if we are still
# playing a track which might not be in the new list. Also if we go to
# a new block, we output the audio URL
if cached:
    debug("Cached end_event %s new %s"%(cached['end_event'], newjsd['end_event']))

if elapsedms >= 0 and endEvent:
    debug("rp-get-flac: using cached data")
    jsd = cached
else:
    debug("outputting audio url because using new metadata or not playing")
    jsd = newjsd
    elapsedms = -1
    newcache(jsd)

# Find the current song in the block based on the elased time    
currentk = None
if elapsedms > 0:
    songs = jsd['song']
    for k in sorted(songs.keys()):
        startms = songs[k]['elapsed']
        endms  = startms + songs[k]['duration']
        debug("k %s Startms %d endms %d event %s" % (k, startms, endms, songs[k]['event']))
        if elapsedms >= startms and elapsedms <= endms:
            currentk = k
            break

if not currentk:
    # Not found ?? Try to reset the thing
    debug(("outputting audio url because current elapsed %d " + \
          " not found in song array") % elapsedms)
    jsd = newjsd
    newcache(jsd)
    out['reload'] = 1
    writesong(jsd, '0')
else:
    writesong(jsd, currentk)
    reload = int((endms - elapsedms)/1000)
    # Last song: reload a bit earlier so that we can queue the URL
    if currentk == len(songs) -1 and reload > 3:
        reload -= 2
    out['reload'] = reload

if PY3:
    sys.stdout.buffer.write(makebytes(json.dumps(out)))
else:
    print("%s" % makebytes(json.dumps(out)))



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class PlayedAlbum:

    def __init__(self):
        self._album_id: str = None
        self._album_played_counter: float = None

    @property
    def album_id(self) -> str:
        return self._album_id

    @album_id.setter
    def album_id(self, value: str):
        self._album_id: str = value

    @property
    def album_played_counter(self) -> float:
        return self._album_played_counter

    @album_played_counter.setter
    def album_played_counter(self, value: float):
        self._album_played_counter: float = value



#
# Copyright (C) 2017-2020 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Manage the tags sections of the tree.
#
# Object Id prefixes:
#     0$uprcl$=[tagname]
#     0$uprcl$albums
#     0$uprcl$items
# Object Ids inside the subsections: complicated

import sys
import os
import sqlite3
import time
import tempfile
import time

from upmplgutils import uplog, direntry
import uprclutils
from uprclutils import rcldoctoentry, cmpentries, cmpitems
import uprclinit
import uprcltagscreate
from uprcltagscreate import _clid, _junctb, recolltosql

_albumclass = "object.container.album.musicAlbum"
_artistclass = "object.container.person.musicArtist"
_completealbum = ">> Complete Album"

def _tblst(tabnames):
    return ",".join(tabnames)


# The browseable object which defines the tree of tracks organized by tags.
class Tagged(object):
    # We maintain a cache for the total count of query statements
    # which are expensive to compute when we are returning a partial
    # slice (which we currently only do when displaying the top-level
    # items list, so the cache can actually never have more than 1 element...)
    _stmt_cnt_cachesize = 20

    def __init__(self, folders, httphp, pathprefix):
        self._httphp = httphp
        self._pprefix = pathprefix
        self._conn = None
        self._init_sqconn()
        self._stmt_cnt_cache = {}
        self._stmt_cnt_cachequeue = []
        self.hidden = []
        recolltosql(self._conn, folders)

    def _init_sqconn(self):
        # We use a separate thread for building the db to ensure
        # responsiveness during this phase.  :memory: handles normally
        # can't be shared between threads, and different :memory: handles
        # access different dbs. The following would work, but it needs
        # python 3.4
        # self._conn = sqlite3.connect('file:uprcl_db?mode=memory&cache=shared')
        # As we can guarantee that 2 threads will never access the db at
        # the same time (the init thread just goes away when it's done),
        # we just disable the same_thread checking on :memory:
        if self._conn is None:
            self._conn = sqlite3.connect(":memory:", check_same_thread=False)
            # Note: it would seem a good thing to disable autocommits, but this actually makes
            # db create performance very slightly worse.
            # self._conn.execute("BEGIN TRANSACTION")
            
    # Create our top-level directories, with fixed entries, and stuff
    # from the tags tables. This may be called (indirectly) from the folders
    # hierarchy, with a path restriction
    def rootentries(self, pid, path=""):
        # uplog("rootentries: pid %s path %s" % (pid, path))
        entries = []
        nalbs = self._albcntforfolder(path)
        entries.append(direntry(pid + "albums", pid, nalbs + " albums"))
        if path:
            where = " WHERE tracks.path LIKE ? "
            args = (path + "%",)
        else:
            where = " "
            args = ()
        c = self._conn.cursor()
        stmt = f"SELECT COUNT(*) from tracks {where}"
        c.execute(stmt, args)
        nitems = str(c.fetchone()[0])
        entries.append(direntry(pid + "items", pid, nitems + " items"))
        subqs = self._subtreetags(
            where,
            [
                "tracks",
            ],
            args,
        )
        tagdisplaytag = uprcltagscreate.getTagDisplayTag()
        for tt in subqs:
            entries.append(direntry(pid + "=" + tt, pid, tagdisplaytag[tt]))
        return entries

    # List all tags which still have multiple values inside this selection level
    def _subtreetags(self, where, seltables, values):
        where = where.strip()
        # uplog(f"_subtreetags: where: [{where}]")
        c = self._conn.cursor()
        tags = []
        indextags = uprcltagscreate.getIndexTags()
        tagtotable = uprcltagscreate.getTagToTable()
        if where:
            conjunct = "AND"
        else:
            conjunct = "WHERE"
        for tt in indextags:
            tb = tagtotable[tt]
            if _junctb(tb) in seltables:
                continue
            stmt = (
                f"SELECT COUNT(DISTINCT {_junctb(tb)}.{_clid(tb)}) "
                f"FROM {_tblst(seltables)},{_junctb(tb)} {where} "
                f"{conjunct} tracks.docidx = {_junctb(tb)}.docidx"
            )
            # uplog(f"subtreetags: executing: {stmt}. Values: {values}")
            c.execute(stmt, values)
            cnt = c.fetchone()[0]
            # uplog(f"subtreetags: {cnt} values for {tb} ({stmt[:200]},{values})")
            if cnt > 1:
                tags.append(tt)
        return tags

    def _stmt_total(self, stmt, values):
        try:
            total = self._stmt_cnt_cache[stmt]
        except:
            c = self._conn.cursor()
            c.execute(stmt, values)
            ids = [r[0] for r in c]
            total = len(ids)
            self._stmt_cnt_cache[stmt] = total
            self._stmt_cnt_cachequeue.append(stmt)
            if len(self._stmt_cnt_cachequeue) > self._stmt_cnt_cachesize:
                del self._stmt_cnt_cache[self._stmt_cnt_cachequeue[0]]
                self._stmt_cnt_cachequeue = self._stmt_cnt_cachequeue[1:]
        return total

    # Build a list of track directory entries for an SQL statement
    # which selects docidxs (SELECT docidx,... FROM tracks WHERE...)
    #
    # NOTE: setting offset or cnt currently precludes sorting, which
    # is unavoidable if we don't cache the whole result (which would
    # use a lot of memory for big collections, around 100MB for 10k
    # songs, significant on a small SBC). At the moment, we only do
    # this for the full top-level items list, which is not really
    # useful anyway.
    def _trackentriesforstmt(self, stmt, values, pid, offset=0, count=0, key=cmpentries):
        # uplog("trackentries: offset %d count %d" % (offset, count))
        total = 0
        if offset != 0 or count != 0:
            total = self._stmt_total(stmt, values)
        if total < 1000:
            # Then return the full set: this is acceptable in terms of
            # performances, and will allow the result to be sorted.
            # plgwithslave knows how to deal with a full set returned
            # from a partial request.
            count = 0
            offset = 0
        if count != 0:
            stmt += " LIMIT %d " % count
        if offset != 0:
            stmt += " OFFSET %d " % offset
        rcldocs = uprclinit.getTree("folders").rcldocs()
        c = self._conn.cursor()
        c.execute(stmt, values)
        entries = [
            rcldoctoentry(pid + "$i" + str(r[0]), pid, self._httphp, self._pprefix, rcldocs[r[0]])
            for r in c
        ]
        # uplog("trackentries: stmt returns %d entries" % len(entries))
        if offset != 0 or count != 0:
            return (offset, total, entries)
        else:
            return sorted(entries, key=key)

    # Return a list of trackids as selected by the current path <selwhere>
    # <values> holds the corresponding values
    def _docidsforsel(self, selwhere, seltables, values):
        c = self._conn.cursor()
        stmt = f"SELECT tracks.docidx FROM {_tblst(seltables)}  {selwhere}  ORDER BY trackno"
        # uplog(f"docidsforsel: executing {stmt}. Values: {values}")
        c.execute(stmt, values)
        return [r[0] for r in c.fetchall()]

    # Expand multiple possibly merged albums to real ones. The tracks
    # always refer to the raw albid, so this is necessary to compute a
    # track list. Multiple albums as input, no sorting.
    def _albids2rawalbids(self, albids):
        c = self._conn.cursor()
        rawalbids = []
        for albid in albids:
            c.execute("""SELECT album_id FROM albums WHERE albalb = ?""", (albid,))
            rows = c.fetchall()
            if len(rows):
                for r in rows:
                    rawalbids.append(r[0])
            else:
                rawalbids.append(albid)
        return rawalbids

    # Expand single possibly merged album into list of ids for component discs
    def _albid2rawalbidssorted(self, albid):
        c = self._conn.cursor()
        c.execute("""SELECT album_id FROM albums WHERE albalb = ? ORDER BY albtdisc""", (albid,))
        rows = c.fetchall()
        if len(rows) <= 1:
            return (albid,)
        else:
            return [r[0] for r in rows]

    # Translate albids so that the ones which are part of a merged
    # album become the merged id. The returned list is same size or
    # smaller because there maybe duplicate merged ids. Used to show
    # album lists
    def _rawalbids2albids(self, rawalbids):
        albids = set()
        c = self._conn.cursor()
        for rawalbid in rawalbids:
            c.execute("""SELECT album_id, albalb FROM albums WHERE album_id = ?""", (rawalbid,))
            alb = c.fetchone()
            if alb[1]:
                albids.add(alb[1])
            else:
                albids.add(alb[0])
        #
        return [id for id in albids]

    # Count albums under file system path. We use albalb because
    # merged albums may come from multiple folders, and have no
    # albfolder. So this returns merged albums for which at least one
    # disk is under this folder path
    def _albcntforfolder(self, path):
        c = self._conn.cursor()
        if path:
            stmt = """SELECT COUNT(DISTINCT albalb) FROM ALBUMS
            WHERE albfolder LIKE ?"""
            args = (path + "%",)
        else:
            stmt = "SELECT COUNT(*) FROM albums WHERE albtdisc is NULL"
            args = ()
        # uplog("_albcntforfolder: stmt %s args %s" % (stmt, args))
        c.execute(stmt, args)
        return str(c.fetchone()[0])

    # Track list for possibly merged album: get tracks from all
    # components, then renumber trackno
    def _trackentriesforalbum(self, albid, pid):
        albids = self._albid2rawalbidssorted(albid)
        # uplog("_trackentriesforalbid: %d -> %s" % (albid, albids))
        # I don't see a way to use a select..in statement and get the
        # order right
        tracks = []
        for albid in albids:
            stmt = """SELECT docidx FROM tracks
            WHERE album_id = ? ORDER BY trackno"""
            tracks += self._trackentriesforstmt(stmt, (albid,), pid)

        tno = None
        for track in tracks:
            tn = 1
            if "upnp:originalTrackNumber" in track:
                tn = int(track["upnp:originalTrackNumber"])
            if tno:
                if tn <= tno:
                    tn = tno + 1
                tno = tn
            else:
                tno = tn
            track["upnp:originalTrackNumber"] = str(tno)
        return tracks

    # Return all albums ids to which any of the selected tracks belong
    def _subtreealbums(self, selwhere, seltables, values):
        stmt = f"SELECT DISTINCT tracks.album_id FROM {_tblst(seltables)} {selwhere}"
        c = self._conn.cursor()
        # uplog(f"subtreealbums: executing {stmt}")
        c.execute(stmt, values)
        rawalbids = [r[0] for r in c]
        albids = self._rawalbids2albids(rawalbids)
        # uplog(f"subtreealbums: returning {albids}")
        return albids

    def _direntriesforalbums(self, pid, where, path=""):
        # uplog("_direntriesforalbums. where: %s" % where)
        c = self._conn.cursor()
        args = (path + "%",) if path else ()
        if path:
            if not where:
                where = """WHERE albfolder LIKE ?"""
            else:
                where += """ AND albfolder LIKE ?"""
            substmt = """SELECT DISTINCT albalb FROM ALBUMS %s""" % where
            where = """WHERE album_id IN (%s)""" % substmt
        else:
            if not where:
                where = """WHERE albtdisc IS NULL"""
            else:
                where += """ AND albtdisc IS NULL"""

        stmt = (
            """SELECT album_id, albtitle, albarturi, albdate, artist.value
        FROM albums LEFT JOIN artist ON artist.artist_id = albums.artist_id
        %s ORDER BY albtitle"""
            % where
        )

        # uplog("_direntriesforalbums: stmt {stmt}")
        c.execute(stmt, args)
        entries = []
        for r in c:
            id = pid + "$" + str(r[0])
            entries.append(
                direntry(
                    id,
                    pid,
                    r[1],
                    arturi=r[2],
                    date=r[3],
                    artist=r[4],
                    upnpclass=_albumclass,
                )
            )
        return entries

    # Called when the search finds one of our synthetic album search
    # results. Create a container entry for it
    def direntryforalbid(self, albid, title=None):
        intalbid = int(albid)
        c = self._conn.cursor()
        stmt = """SELECT album_id, albtitle, albarturi, albdate, artist.value
        FROM albums LEFT JOIN artist ON artist.artist_id = albums.artist_id
        WHERE album_id = ? AND albtdisc is NULL  ORDER BY albtitle"""
        args = (intalbid,)
        c.execute(stmt, args)
        for r in c:
            if not title:
                title = r[1]
            pid = uprclinit.getObjPrefix() + "albums"
            id = pid + "$" + albid
            return direntry(
                id,
                pid,
                title,
                arturi=r[2],
                date=r[3],
                artist=r[4],
                upnpclass=_albumclass,
            )
        return None

    # Called when the recoll search finds an artist record result.
    def direntryforartid(self, artid):
        intartid = int(artid)
        c = self._conn.cursor()
        stmt = "SELECT value FROM artist WHERE artist_id = ?"
        args = (int(artid),)
        c.execute(stmt, args)
        for r in c:
            pid = uprclinit.getObjPrefix() + "=Artist"
            id = pid + "$" + artid
            return direntry(id, pid, r[0], upnpclass=_artistclass)
        return None

    # This is called when an 'albums' element is encountered in the
    # selection path. i is the index of the albums element. The tree under
    # albums has a well defined structure: ql=len(qpath), we have an
    # albums list if i is the last element (i == ql-1), an album track
    # list for i == ql-2 (we then have an albid at ql-1), and a 'Complete
    # album' query if i == ql-3 (...$albums$xxx$showca)
    def _tagsbrowsealbums(self, pid, qpath, i, selwhere, seltables, values):
        # uplog(f"_tagsbrowsealbums: pid {pid} qpath {qpath} i {i} where {selwhere} values {values}"
        c = self._conn.cursor()
        entries = []
        if i == len(qpath) - 1:
            # List of albums to which belong any track from selection
            albidsl = self._subtreealbums(selwhere, seltables, values)
            albids = ",".join([str(a) for a in albidsl])
            where = " WHERE album_id in (" + albids + ") "
            entries = self._direntriesforalbums(pid, where)
        elif i == len(qpath) - 2:
            # Album track list. Maybe a merged album->multiple phys albids
            albid = int(qpath[-1])
            rawalbids = self._albids2rawalbids((albid,))
            # uplog(f"_tagsbrowsealbums: albid {albid} rawalbids {rawalbids}")
            stmt = """SELECT COUNT(docidx) FROM tracks
                WHERE album_id IN (%s)""" % ",".join(
                "?" * len(rawalbids)
            )
            c.execute(stmt, rawalbids)
            r = c.fetchone()
            ntracks = int(r[0])
            docidsl = self._docidsforsel(selwhere, seltables, values)
            stmt = """SELECT docidx FROM tracks 
                WHERE album_id IN (%s) AND docidx IN (%s)""" % (
                ",".join("?" * len(rawalbids)),
                ",".join("?" * len(docidsl)),
            )
            entries = self._trackentriesforstmt(stmt, rawalbids + docidsl, pid)
            if ntracks != len(entries):
                id = pid + "$" + "showca"
                entries = [self.direntryforalbid(str(albid), title=_completealbum)] + entries
        elif i == len(qpath) - 3:
            # 'Complete album' entry
            # Note that minim has an additional level here, probably to
            # present groups or multiple groups ? The trackids ids are
            # like:
            #    0$=Composer$17738$albums$2$showca.0$hcalbum$*i13458
            # I don't know what the .0 is for.
            # The 'hcalbum' level usually has 2 entries '>> Hide Content'
            # and the album title. TBD the way we do it currently, we lose the album title in the
            # tracks screen header (it's "Complete Album")
            albid = int(qpath[-2])
            entries = self._trackentriesforalbum(albid, pid)

        return entries

    # This is called when an 'items' element is encountered in the selection path.
    def _tagsbrowseitems(self, pid, qpath, i, selwhere, seltables, values):
        stmt = f"SELECT tracks.docidx FROM {_tblst(seltables)} {selwhere}"
        c = self._conn.cursor()
        rows = c.execute(stmt, values)
        docids = [r[0] for r in rows]
        albids = self._subtreealbums(selwhere, seltables, values)
        entries = []
        displaytracks = True
        if len(albids) == 1:
            # Only display '>> Complete album' if not all tracks
            # already there. If all tracks are there, we display
            # the album entry (with the same id value: show album)
            albid = albids[0]
            tlist = self._trackentriesforalbum(albid, pid)
            # Replace $items with $albums for the album entry
            id = pid.replace("$items", "$albums") + f"${albid}$showca"
            if len(tlist) != len(docids):
                entries.append(self.direntryforalbid(str(albid), title=_completealbum))

            # We used to show an album entry here, but the album was probably already shown at the
            # level above (because there is only one), and it's better to show the tracks in title
            # order. Kept around because I'm not sure that there are not cases where we'd want it
            # anyway.
            # else:
            #    displaytracks = False
            #    el = self._direntriesforalbums(pid, f"WHERE album_id = {albid}")
            #    el[0]['id'] = id
            #    entries.append(el[0])
        if displaytracks:
            rcldocs = uprclinit.getTree("folders").rcldocs()
            entries += sorted(
                [
                    rcldoctoentry(
                        pid + "$i" + str(docid), pid, self._httphp, self._pprefix, rcldocs[docid]
                    )
                    for docid in docids
                ],
                key=cmpitems,
            )
        return entries

    # Main browsing routine. Given an objid, translate it into a select
    # statement, plus further processing, and return the corresponding
    # records
    def _tagsbrowse(self, pid, qpath, flag, path=""):
        # uplog(f"tagsbrowse. pid {pid} qpath {qpath}")

        tagdisplaytag = uprcltagscreate.getTagDisplayTag()

        # Walk the qpath, which was generated from the objid and
        # defines what tracks are selected and what we want to
        # display. E.g =Artist$21$=Date would display all distinct
        # dates for tracks by Artist #21. =Artist$21$=Date$48 the data
        # for date 48 (the numbers are indexes into the aux tables)
        qlen = len(qpath)
        selwhat = ""
        seltables = [
            "tracks",
        ]
        if path:
            selwhere = " WHERE tracks.path LIKE ? "
            values = [
                path + "%",
            ]
        else:
            selwhere = ""
            values = []
        tagtotable = uprcltagscreate.getTagToTable()
        i = 0
        while i < qlen:
            elt = qpath[i]

            # Detect the special values: albums items etc. here. Their
            # presence changes how we process the rest (showing tracks and
            # albums and not dealing with other tags any more)
            if elt == "albums":
                return self._tagsbrowsealbums(pid, qpath, i, selwhere, seltables, values)
            elif elt == "items":
                return self._tagsbrowseitems(pid, qpath, i, selwhere, seltables, values)

            # '=colname'. Set the current column name, which will be used
            # in different ways depending if this is the last element or
            # not.
            if elt.startswith("="):
                col = tagtotable[elt[1:]]

            selwhere = selwhere + " AND " if selwhere else " WHERE "
            seltables.append(_junctb(col))
            if i == qlen - 1:
                # We can only get here if the qpath ends with '=colname'
                # (otherwise the else branch below fetches the 2 last
                # elements and breaks the loop). We want to fetch all
                # unique values for the column inside the current selection.

                # e.g. artist.artist_id, artist.value
                selwhat = f"{col}.{_clid(col)}, {col}.value"
                # e.g. tracks.artist_id = artist.artist_id
                selwhere += (
                    f"tracks.docidx = {_junctb(col)}.docidx AND "
                    f"{_junctb(col)}.{_clid(col)} = {col}.{_clid(col)}"
                )
            else:
                # Look at the value specified for the =xx column. The
                # selwhat value is only used as a flag
                selwhat = "tracks.docidx"
                selwhere += (
                    f"tracks.docidx = {_junctb(col)}.docidx AND " f"{_junctb(col)}.{_clid(col)} = ?"
                )
                i += 1
                values.append(int(qpath[i]))
            i += 1

        entries = []
        if selwhat == "tracks.docidx":
            # uplog(f"tagsbrowse: showing remaining multivalued tags")
            # We are displaying content for a given value of a given tag
            docids = self._docidsforsel(selwhere, seltables, values)
            albids = self._subtreealbums(selwhere, seltables, values)
            subqs = self._subtreetags(selwhere, seltables, values)
            displaytracks = True
            if len(albids) > 1:
                id = pid + "$albums"
                label = "%d albums"
                entries.append(direntry(id, pid, label % len(albids)))
            elif len(albids) == 1:
                # Only display '>> Complete album' if not all tracks
                # already there. If all tracks are there, we display
                # the album entry (with the same id value: show album)
                albid = albids[0]
                tlist = self._trackentriesforalbum(albid, pid)
                id = pid + "$albums$" + str(albid) + "$showca"
                if len(tlist) != len(docids):
                    entries.append(self.direntryforalbid(str(albid), title=_completealbum))
                else:
                    displaytracks = False
                    el = self._direntriesforalbums(pid, "WHERE album_id = %s" % albid)
                    el[0]["id"] = id
                    entries.append(el[0])

            if subqs:
                id = pid + "$items"
                label = "%d items"
                entries.append(direntry(id, pid, label % len(docids)))
                for tt in subqs:
                    id = pid + "$=" + tt
                    entries.append(direntry(id, pid, tagdisplaytag[tt]))
            elif displaytracks:
                rcldocs = uprclinit.getTree("folders").rcldocs()
                tracks = []
                for docidx in docids:
                    id = pid + "$*i" + str(docidx)
                    tracks.append(
                        rcldoctoentry(id, pid, self._httphp, self._pprefix, rcldocs[docidx])
                    )
                entries += sorted(tracks, key=cmpitems)
        else:
            # Showing all values at this point for given column
            # SELECT col.col_id, col.value FROM tracks, col
            # WHERE tracks.col_id = col.col_id
            # GROUP BY tracks.col_id
            # ORDER BY col.value
            stmt = (
                f"SELECT {selwhat} FROM {_tblst(seltables)}, {col} "
                f"{selwhere} GROUP BY {col}.value ORDER BY {col}.value"
            )
            # uplog(f"tagsbrowse: showing values for {selwhat} : <{stmt}> values {values}")
            c = self._conn.cursor()
            c.execute(stmt, values)
            for r in c:
                id = pid + "$" + str(r[0])
                entries.append(direntry(id, pid, r[1]))
        return entries

    # Browse the top-level tree named like 'xxx albums'. There are just 2
    # levels: the whole albums list, then for each entry the specified
    # albums track list
    def _albumsbrowse(self, pid, qpath, flag, path=""):
        c = self._conn.cursor()
        entries = []
        if len(qpath) == 1:
            entries = self._direntriesforalbums(pid, "", path)
        elif len(qpath) == 2:
            e1 = qpath[1]
            album_id = int(e1)
            entries = self._trackentriesforalbum(album_id, pid)
        else:
            raise Exception("Bad path in album tree (too deep): <%s>" % qpath)

        return entries

    # Implement the common part of browse() and browseFolder()
    def _dobrowse(self, pid, flag, qpath, folder="", offset=0, count=0):
        # uplog(f"Tags:_dobrowse: pid {pid} qpath {qpath} folder [{folder}] ofs {offset} cnt {count}")
        if qpath[0] == "items":
            args = (folder + "%",) if folder else ()
            folderwhere = " WHERE tracks.path LIKE ? " if folder else " "
            stmt = "SELECT docidx FROM tracks" + folderwhere
            entries = self._trackentriesforstmt(stmt, args, pid, offset, count, key=cmpitems)
        elif qpath[0] == "albums":
            entries = self._albumsbrowse(pid, qpath, flag, folder)
        elif qpath[0].startswith("="):
            entries = self._tagsbrowse(pid, qpath, flag, folder)
        else:
            raise Exception(f"Bad path in tags tree (start): <{qpath}>")
        return entries

    # Call from the folders tree when Tag View is selected. Reproduces
    # the general view by tags, but with selection for only the files
    # under the specified folder.
    #
    # Somewhat like minim: under ">> Tag View", 2 entries, ">> Hide
    # Contents/" and simplename/ (last elt of folder path). If ">> Hide
    # Contents/" is visited, like would happen with a CP recursive
    # walk, simplename/ is marked for 2 S and will be empty if
    # visited. If ">> Hide Contents/" is not visited, or after 2 S,
    # simplename/ will contain the local tags subtree.
    #
    def browseFolder(self, pid, flag, pthremain, folder):
        # uplog(f"Tags:browseFolder: objpath {pthremain} folder {folder}")
        l = pthremain.split("$")
        # 1st elt in list is empty because pthremain begins with $. so
        # len(l)==2 is the root of tags from this folder
        entries = []
        if len(l) < 2:
            # ??
            return entries
        if len(l) == 2:
            # (0$uprcl$folders$d2)$tagview.0
            entries.append(direntry(pid + "$hchide", pid, ">> Hide Contents"))
            entries.append(direntry(pid + "$hctags", pid, uprclutils.basename(folder)))
            return entries
        elif len(l) == 3:
            ppid = "$".join(pid.split("$")[:-1])
            value = l[-1]
            if value == "hchide":
                # (0$uprcl$folders$d2)$tagview.0$hchide
                self.hidden.insert(0, (time.time(), ppid))
                return entries
            else:
                # (0$uprcl$folders$d2)$tagview.0$hctags
                now = time.time()
                for i in range(len(self.hidden)):
                    # uplog("Browsefolder: hidden: since %d pid %s" %
                    #      (int(now - self.hidden[i][0]), self.hidden[i][1]))
                    if self.hidden[i][0] <= now - 2:
                        self.hidden = self.hidden[:i]
                        break
                    elif self.hidden[i][1] == ppid:
                        return entries
                return self.rootentries(pid + "$", folder)
        else:
            # (0$uprcl$folders$d2)$tagview.0$hctags$=Artist...
            # Back to normal browsing, restricted to entries occuring under the given folder
            qpath = l[3:]
            return self._dobrowse(pid, flag, qpath, folder)

    def _browsemeta(self, objid):
        # very few control points use this, and it is useless because we return the full metadata in
        # the container entries. This would be too complicated to implement in the tags tree, so we
        # only do it for items, and in a very brutal way, by browsing the parent and finding the
        # entry in the result.
        # uplog(f"_browsemeta: objid {objid}")
        anchorlen = 2
        ipos = objid.rfind("$i")
        if ipos == -1:
            ipos = objid.rfind("$*i")
            if ipos == -1:
                return []
            else:
                anchorlen = 3
        if objid[ipos:] == "$items":
            return []
        id = objid[ipos + anchorlen :]
        try:
            iid = int(id)
        except:
            return []
        pid = objid[:ipos]
        entries = self.browse(pid, "children", 0, 0)
        for e in entries:
            if e["id"] == objid:
                return [
                    e,
                ]
        return []

    # Top level browse routine. Handle the special cases and call the
    # appropriate worker routine. idpath is something like 0$uprcl$=tagname$tagvalue...
    def browse(self, pid, flag, offset, count):
        if flag == "meta":
            return self._browsemeta(pid)
        idpath = pid.replace(uprclinit.getObjPrefix(), "", 1)
        # Idpath now looks like =Artist$14$=Genre...
        uplog("tags:browse: idpath <%s>" % idpath)
        qpath = idpath.split("$")
        return self._dobrowse(pid, flag, qpath, offset=offset, count=count)


############ Misc test/trial code, not used by uprcl ########################


def misctries():
    c = self._conn.cursor()
    c.execute("""SELECT COUNT(*) FROM tracks""")
    uplog("Count(*) %d" % (c.fetchone()[0],))

    # for row in c.execute('''SELECT album
    #                        FROM tracks where artist LIKE "%Gould%"
    #                        GROUP BY album'''):
    #    uplog("%s" % (row,))

    # For some strange reason it appears that GROUP BY is faster than SELECT
    # DISTINCT
    stmt = """SELECT album FROM tracks GROUP BY album ORDER BY album"""
    start = timer()
    for row in c.execute(stmt):
        # uplog("%s" % (row[0].encode('UTF-8')))
        pass
    end = timer()
    uplog("Select took %.2f Seconds" % (end - start))
    for row in c.execute("""SELECT COUNT(DISTINCT album) from tracks"""):
        uplog("Album count %d" % row[0])


if __name__ == "__main__":
    confdir = "/home/dockes/.recoll-mp3"
    from recoll import recoll

    def fetchalldocs(confdir):
        allthedocs = []
        rcldb = recoll.connect(confdir=confdir)
        rclq = rcldb.query()
        rclq.execute("mime:*", stemming=0)
        uplog("Estimated alldocs query results: %d" % (rclq.rowcount))
        maxcnt = 1000
        totcnt = 0
        while True:
            docs = rclq.fetchmany()
            for doc in docs:
                allthedocs.append(doc)
                totcnt += 1
            if (maxcnt > 0 and totcnt >= maxcnt) or len(docs) != rclq.arraysize:
                break
        uplog("Retrieved %d docs" % (totcnt,))
        return allthedocs

    start = timer()
    docs = fetchalldocs(confdir)
    end = timer()
    uplog("Recoll extract took %.2f Seconds" % (end - start))
    start = timer()
    recolltosql(docs)
    end = timer()
    uplog("SQL db create took %.2f Seconds" % (end - start))



#!/usr/bin/env python3
#
# Copyright (C) 2023 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import json
import re
import cmdtalkplugin
import conftree
from html import escape as htmlescape, unescape as htmlunescape

import pyradios
from pyradios import facets

from urllib.request import urlopen
import re as r

from upmplgutils import uplog, setidprefix, direntry, getOptionValue

# Prefix for object Ids. This must be consistent with what contentdirectory.cxx does
__plugin_name = "radio-browser"
_g_myprefix = f"0${__plugin_name}$"
setidprefix(__plugin_name)

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)


def _get_public_ip() -> str:
    d = str(urlopen('http://checkip.dyndns.com/').read())
    return r.compile(r'Address: (\d+\.\d+\.\d+\.\d+)').search(d).group(1)


def _get_country_code() -> str:
    cfg_country = getOptionValue(f"{__plugin_name}country")
    if cfg_country: return cfg_country
    token = getOptionValue(f"{__plugin_name}ipinfotoken")
    if not token: return None
    public_ip = _get_public_ip()
    if not public_ip: return None
    url = f"https://www.ipinfo.io/{public_ip}?token={token}"
    # call the API and save the response
    with urlopen(url) as response:
        response_content = response.read()
    # parsing the response 
    data = json.loads(response_content)
    country = data["country"] if "country" in data else None
    msgproc.log(f"country: [{country}]")
    return country
    

_g_rb = None
_g_country = None
def _initradios():
    global _g_rb
    global _g_country
    if not _g_rb:
        _g_rb = pyradios.RadioBrowser()
        global _ccodetoc
        _ccodetoc = {}
        countries = _g_rb.countries()
        for c in countries:
            _ccodetoc[c["iso_3166_1"]] = c["name"]
    if not _g_country:
        _g_country = _get_country_code()

def _radiotoentry(pid, id, radio):
    """Translate radio record from radio-browser into plugin output item"""
    #uplog(f"radioToEntry: pid {pid} id {id} radio {radio}")
    # if this comes from a 'browse meta', the id is set, else compute it
    if id is None:
        objid = pid
        if objid[-1] != "$":
            objid += "$"
        objid += str(radio["stationuuid"])
    else:
        objid = id

    mime = "audio/mpeg"
    if "codec" in radio:
        if radio["codec"].lower() == "flac":
            mime = "audio/flac"
        elif radio["codec"].lower().find("aac") == 0:
            mime = "audio/aac"
        elif radio["codec"].lower() == "ogg":
            mime = "audio/ogg"
    else:
        mime = "audio/mpeg"
    entry = {
        'pid': pid,
        'id': objid,
        'uri': radio["url_resolved"],
        'tp': 'it',
        'res:mime': mime,
        'upnp:class': 'object.item.audioItem.audioBroadcast',
        'upnp:album': radio["codec"] if "codec" in radio else "Unknown",
        'tt': radio["name"],
        # This is for Kodi mostly, to avoid displaying a big "Unknown" placeholder
        'upnp:artist': "Internet Radio"
    }
    if "favicon" in radio:
        entry["upnp:albumArtURI"] = radio["favicon"]
    return entry


@dispatcher.record('trackuri')
def trackuri(a):
    # We generate URIs which directly point to the radio, so this method should never be called.
    raise Exception("trackuri: should not be called for radio-browser!")


def _returnentries(entries):
    """Build plugin browse or search return value from items list"""
    return {"entries" : json.dumps(entries), "nocache" : "0"}


_crittotitle_a = {"tags": "tags", "countrycodes": "countries", "languages": "languages"}
def _crittotitle(crit):
    """Translate filtering field name to displayed title"""
    return _crittotitle_a[crit]

def _pathtocrits(path):
    if path:
        lpath = path.split("/")
    else:
        lpath = []

    # Compute the filtering criteria from the path
    # All field names we see: used to restrict those displayed at next step
    setcrits = []
    # Arguments to the facets filtering object: tagname,value pairs
    filterargs = {}
    # Final values after walking the path. Decide how/what to display next
    crit = None
    value = None
    argtrans = {"tags": "tag", "countrycodes": "countrycode", "languages": "language"}
    for idx in range(len(lpath)):
        if idx & 1:
            continue
        crit = lpath[idx]
        if idx < len(lpath)-1:
            value = htmlunescape(lpath[idx+1])
            setcrits.append(crit)
            filterargs[argtrans[crit]] = value
        else:
            value = None
            break
    return crit, value, setcrits, filterargs

def _objidtopath(objid):
    if objid.find(_g_myprefix) != 0:
        raise Exception(f"radio-browser: bad objid {objid}: bad prefix")
    return objid[len(_g_myprefix):].lstrip("/")
    
@dispatcher.record('browse')
def browse(a):
    msgproc.log(f"browse: args: --{a}--")
    _initradios()
    if 'objid' not in a:
        raise Exception("No objid in args")

    objid = a['objid']
    path = _objidtopath(objid)

    entries = []
    bflg = a['flag'] if 'flag' in a else 'children'
    if bflg == 'meta':
        # Only Kodi does this.
        uuidpos = path.find("$")
        if uuidpos >= 0:
            uuid = path[uuidpos+1:]
            radios = _g_rb.station_by_uuid(uuid)
            entries = [_radiotoentry(objid, None, radios[0]),]
        else:
            entries = [direntry(id, objid, "forkodi"),]
        return _returnentries(entries)

    lastcrit, lastvalue, setcrits, filterargs = _pathtocrits(path)

    # A None lastcrit can only happen if the path was empty (root). Else lastvalue is empty if we
    # want to display the possible values for lastcrit, else it's the filtering value (which we only
    # use as part of filterargs). setcrits is the list of already frozen criteria (don't display
    # them any more), and filterargs is the current state of the selection. setcrits and
    # filterargs.keys() are equivalent may use different keys (e.g. countries vs countrycodes).
    if lastcrit:
        filter = facets.RadioFacets(_g_rb, **filterargs)
    else:
        filter = None

    if lastcrit == "radios":
        # A path ending with "radios" means that we want the list of possible radios matching the
        # filter.
        for radio in filter.result:
            entries.append(_radiotoentry(objid, None, radio))
        return _returnentries(entries)

    if lastvalue is None and lastcrit is not None:
        # The path ended with a tagname. Get the list of possible values.
        if lastcrit == "countrycodes":
            codes = [entry["name"] for entry in filter.countrycodes
                     if entry["name"] and entry["name"] in _ccodetoc]
            # Sort alphabetically for countries, popularity does not really make sense.
            values = sorted([(code, _ccodetoc[code]) for code in codes], key=lambda i: i[1])
        elif lastcrit == "languages":
            values = [(entry["name"], entry["name"]) for entry in filter.languages]
        elif lastcrit == "tags":
            values = [(entry["name"], entry["name"]) for entry in filter.tags]
        else:
            raise(Exception(f"Bad objid {objid} (bad tagname [{lastcrit}])"))
        for value, title in values:
            if value:
                id = objid + "/" + htmlescape(value, quote=True)
                entries.append(direntry(id, objid, title))
    else:
        # Path is root or ends with tag value. List the remaining tagnames if any.
        for tagname in ("countrycodes", "languages", "tags"):
            if not tagname in setcrits:
                id = objid + "/" + tagname
                entries.append(direntry(id, objid, _crittotitle(tagname)))
        # add a shortcut to local radio
        # if _g_country is available
        if _g_country and (not setcrits or len(setcrits) == 0):
            id = objid + "/countrycodes/" + _g_country
            entries.append(direntry(id, objid, "local"))

    # Link to the available radios at this stage. In root, we don't build the filter, so don't have
    # data here at the top level (don't want to show 30k radios).
    if filter and lastvalue:
        entries.append(direntry(objid + "/radios", objid, f"{len(filter.result)} radios"))

    # msgproc.log(f"browse: returning --{entries}--")
    return _returnentries(entries)


@dispatcher.record('search')
def search(a):
    msgproc.log("search: [%s]" % a)
    objid = a["objid"]
    path = _objidtopath(objid)
    lastcrit, lastvalue, setcrits, filterargs = _pathtocrits(path)
    _initradios()
    sstr = a["value"]
    filterargs["name"] = sstr
    result = _g_rb.search(**filterargs)
    #uplog(f"SEARCH RESULT for [{sstr}]: {result}")
    entries = []
    for radio in result if result else []:
        entries.append(_radiotoentry(objid, None, radio))
    return _returnentries(entries)


msgproc.mainloop()



# Copyright (C) 2019-2020 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import conftree
from uprclutils import uplog


class MinimConfig(object):
    def __init__(self, fn=""):
        if fn:
            try:
                with open(fn, "rb") as f:
                    pass
            except Exception as ex:
                uplog(f"MinimConfig: could not open/read: {ex}")
                fn = "/dev/null"
            self.conf = conftree.ConfSimple(fn)
        else:
            self.conf = conftree.ConfSimple("/dev/null")
        self.quotes = "\"'"
        self.escape = ""
        self.whitespace = ", "

    def getsimplevalue(self, nm):
        s = self.conf.get(nm)
        if s:
            return s.strip()
        else:
            return s

    def getboolvalue(self, nm, dflt):
        val = self.getsimplevalue(nm)
        if val is None or val == "":
            return dflt
        if val == "0" or val.lower()[0] == "f" or val.lower()[0] == "n":
            return False
        else:
            return True

    # Split on commas and colons, a common minim format and return as
    # list of pairs
    def minimsplitsplit(self, str):
        out = []
        if not str:
            return out
        # For some reason, the active colons are backslash-escaped in
        # the file
        lst = str.replace("\\:", ":").split(",")
        for e in lst:
            l = e.split(":")
            if len(l) == 0:
                a = ""
                b = ""
            elif len(l) == 1:
                a = l[0]
                b = ""
            else:
                a = l[0]
                b = l[1]
            out.append((a.strip(), b.strip()))
        return out

    # Comma-separated list. Each element maybe a simple tagname.option
    # value or an \= assignement, with a comma-separated list of values in
    # braces. Returns a list of quadruplet: (tag, neg, optname, values)
    def minimsplitbraceslist(self, value):
        # Split everything on commas
        l1 = value.split(",")
        # Re-join parts inside braces, creating list of options
        inbraces = False
        accum = ""
        l2 = []
        for e in l1:
            if inbraces:
                accum += ", " + e
                if e.find("}"):
                    inbraces = False
                    l2.append(accum)
                    accum = ""
                continue
            if e.find("{") != -1 and e.find("}") == -1:
                inbraces = True
                accum = e
                continue
            l2.append(e)
        # Parse each option [-]tag.option={tag1,tag2} or [-]tag.option
        alloptions = []
        for e in l2:
            lhsrhs = e.split("=")
            if len(lhsrhs) == 2:
                values = [v.strip().lower() for v in lhsrhs[1].lstrip("{").rstrip("}").split(",")]
            else:
                values = []
            tagopt = lhsrhs[0].split(".")
            tag = tagopt[0].strip()
            if tag.startswith("-"):
                neg = True
                tag = tag[1:]
            else:
                neg = False
            opt = ".".join(tagopt[1:])
            alloptions.append((tag, neg, opt, values))
        return alloptions

    def getexcludepatterns(self):
        spats = self.conf.get("minimserver.excludePattern")
        if spats:
            lpats = conftree.stringToStrings(
                spats, quotes=self.quotes, escape=self.escape, whitespace=self.whitespace
            )
            spats = conftree.stringsToString(lpats)
        # uplog("skippedNames from Minim excludePattern: %s" % spats)
        return spats

    def gettagvalue(self):
        stagv = self.conf.get("minimserver.tagValue")
        tagvalue = None
        if stagv:
            tagvalue = self.minimsplitbraceslist(stagv)
        return tagvalue

    def getaliastags(self):
        aliases = []
        saliases = self.conf.get("minimserver.aliasTags")
        # uplog("Minim:getaliastags:in: [%s]" % saliases)
        lst = self.minimsplitsplit(saliases)
        for orig, target in lst:
            orig = orig.lower()
            target = target.lower()
            rep = False
            if target.startswith("-"):
                rep = True
                target = target[1:]
            aliases.append((orig, target, rep))
        # uplog("Minim:getaliastags:out: %s" % aliases)
        return aliases

    def getindextags(self):
        indextags = []
        sit = self.conf.get("minimserver.indexTags")
        # uplog("Minim:getindextags:in: [%s]" % sit)
        if sit:
            indextags = self.minimsplitsplit(sit)
        # uplog("Minim:getindextags:out: %s" % indextags)
        return indextags

    def getitemtags(self):
        itemtags = []
        sit = self.conf.get("minimserver.itemTags")
        # uplog("Minim:getitemtags:in: [%s]" % sit)
        if sit:
            itemtags = [i.strip() for i in sit.split(",")]
        # uplog("Minim:getitemtags:out: %s" % itemtags)
        return itemtags

    def getcontentdirs(self):
        # minim uses '\\n' as a separator for directories (actual
        # backslash then n), not newline. Weird...
        cdirs = []
        s = self.conf.get("minimserver.contentDir")
        if s:
            cdirs = s.replace("\\n", "\n").split("\n")
        return cdirs

    def gettranscodingspec(self):
        s = self.conf.get("stream.transcode")
        if not s:
            return None

        specs = self.minimsplitsplit(s)
        # each spec is a pair of (input,output)
        # TBD
        pass



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from track_adapter import TrackAdapter
from played_track import PlayedTrack


class PlayedTrackAdapter(TrackAdapter):

    def __init__(self, track: PlayedTrack):
        self._track: PlayedTrack = track

    def get_id(self) -> str:
        return self._track.track_id

    def get_name(self) -> str:
        return self._track.track_name

    def get_duration(self) -> int:
        return self._track.track_duration

    def get_track_num(self) -> int:
        return self._track.track_num

    def get_volume_num(self) -> int:
        return self._track.volume_num

    def get_album_num_volumes(self) -> int:
        return self._track.album_num_volumes

    def get_album_track_count(self) -> int:
        return self._track.album_track_count

    def get_album_name(self) -> str:
        return self._track.album_name

    def get_album_id(self) -> str:
        return self._track.album_id

    def get_album_artist_name(self) -> str:
        return self._track.album_artist_name

    def get_audio_quality(self) -> str:
        return self._track.audio_quality

    def get_image_url(self) -> str:
        return self._track.image_url

    def explicit(self) -> bool:
        return True if self._track.explicit == 1 else False

    def get_artist_name(self) -> str:
        return self._track.artist_name

    def get_bit_depth(self) -> int:
        return self._track.bit_depth

    def get_sample_rate(self) -> int:
        return self._track.sample_rate



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subsonic_connector.album import Album
from subsonic_connector.album_list import AlbumList
from subsonic_connector.response import Response
import request_cache
import connector_provider
import config
import constants
from msgproc_provider import msgproc
import random


class TagToEntryContext:

    def __init__(self):
        self.__first_load: bool = True
        self.__random_album_list: list[Album] = []

    @property
    def random_album_list(self) -> list[Album]:
        verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
        if verbose:
            msgproc.log("TagToEntryContext::random_album_list enter ...")
        if not self.__random_album_list or len(self.__random_album_list) == 0:
            # load some random albums
            if verbose:
                msgproc.log("TagToEntryContext::random_album_list (loading) ...")
            res: Response[AlbumList] = None
            if self.__first_load:
                if verbose:
                    msgproc.log("TagToEntryContext::random_album_list loading from request cache ...")
                # res = connector_provider.get().getRandomAlbumList(size=config.get_items_per_page())
                res = request_cache.get_random_album_list(size=config.get_items_per_page())
                self.__first_load = False
            else:
                # don't use cache
                if verbose:
                    msgproc.log("TagToEntryContext::random_album_list loading random albums ...")
                res = connector_provider.get().getRandomAlbumList(size=config.get_items_per_page())
            if res and res.isOk():
                # store.
                res_list: list[Album] = res.getObj().getAlbums()
                self.__random_album_list = list(res_list) if res_list else []
                if (len(self.__random_album_list) > 0 and
                        config.get_config_param_as_bool(constants.ConfigParam.ALLOW_SHUFFLE_RANDOM_ALBUM_FOR_FRONT_PAGE_TAGS)):
                    random.shuffle(self.__random_album_list)
            else:
                # storing an empty list
                self.__random_album_list = []
        if verbose:
            msgproc.log(f"TagToEntryContext::random_album_list return a list with [{len(self.__random_album_list)}] items")
        return self.__random_album_list



import base64
import re
from collections import OrderedDict

import requests


class Spoofer:
    def __init__(self):
        self.seed_timezone_regex = (
            r'[a-z]\.initialSeed\("(?P<seed>[\w=]+)",window\.utimezone\.(?P<timezone>[a-z]+)\)'
        )
        # note: {timezones} should be replaced with every capitalized timezone joined by a |
        self.info_extras_regex = r'name:"\w+/(?P<timezone>{timezones})",info:"(?P<info>[\w=]+)",extras:"(?P<extras>[\w=]+)"'
        self.appId_regex = r'production:{api:{appId:"(?P<app_id>\d{9})",appSecret:"(?P<secret>\w{32})"},braze:.\(.\({},.\),{},{apiKey:"([-0-9a-fA-F]{36})"}\),extra:.}'
        login_page_request = requests.get("https://play.qobuz.com/login")
        login_page = login_page_request.text
        bundle_url_match = re.search(
            r'<script src="(/resources/\d+\.\d+\.\d+-[a-z]\d{3}/bundle\.js)"></script>',
            login_page,
        )
        bundle_url = bundle_url_match.group(1)
        bundle_req = requests.get("https://play.qobuz.com" + bundle_url)
        self.bundle = bundle_req.text

    def getAppId(self):
        return re.search(self.appId_regex, self.bundle).group("app_id")

    def getSecrets(self):
        seed_matches = re.finditer(self.seed_timezone_regex, self.bundle)
        secrets = OrderedDict()
        for match in seed_matches:
            seed, timezone = match.group("seed", "timezone")
            secrets[timezone] = [seed]
        """The code that follows switches around the first and second timezone. Why? Read on:
            Qobuz uses two ternary (a shortened if statement) conditions that should always return false.
            The way Javascript's ternary syntax works, the second option listed is what runs if the condition returns false.
            Because of this, we must prioritize the *second* seed/timezone pair captured, not the first.
        """
        keypairs = list(secrets.items())
        secrets.move_to_end(keypairs[1][0], last=False)
        info_extras_regex = self.info_extras_regex.format(
            timezones="|".join([timezone.capitalize() for timezone in secrets])
        )
        info_extras_matches = re.finditer(info_extras_regex, self.bundle)
        for match in info_extras_matches:
            timezone, info, extras = match.group("timezone", "info", "extras")
            secrets[timezone.lower()] += [info, extras]
        for secret_pair in secrets:
            secrets[secret_pair] = base64.standard_b64decode(
                "".join(secrets[secret_pair])[:-44]
            ).decode("utf-8")
        return secrets


if __name__ == "__main__":
    spoofer = Spoofer()
    print("%s" % spoofer.getSecrets())
    print("%s" % spoofer.getAppId())



# defined to emulate the session object from the tidal module, which is
# defined in the tidalapi part (we want to keep the qobuz api/ dir as
# much alone as possible.
from __future__ import print_function

import sys
import json
from upmplgmodels import (
    Artist,
    Album,
    Track,
    Playlist,
    SearchResult,
    Category,
    Genre,
    Model,
)
from upmplgutils import *
import hraapi
import urllib
import datetime
import traceback
import base64


class Session(object):
    def __init__(self):
        self.api = hraapi.HRAAPI()

    def login(self, username, password, lang):
        self.api.setlang(lang)
        data = self.api.user_login(username=username, password=password)
        if data:
            return True
        else:
            return False

    def isloggedin(self):
        return self.api.isloggedin()

    def get_categories(self):
        return list(map(_parse_category, self.api.getAllCategories()))

    def get_genres(self, parent=None):
        data = self.api.getAllGenres()
        return [_parse_genre(g) for g in data]

    def get_catg_albums(self, prefix):
        data = self.api.listCategoryContent(category=prefix)
        return [_parse_album(a) for a in data]

    def get_alluserplaylists(self):
        data = self.api.listAllUserPlaylists()
        if not data:
            return []
        return [_parse_playlist(p) for p in data]

    def get_user_playlist(self, id):
        data = self.api.getUserPlaylist(playlist_id=id)
        if not data:
            return []
        return [_parse_track(t) for t in data]

    def get_alluseralbums(self):
        data = self.api.listAllUserAlbums()
        if not data:
            return []
        return [_parse_album(p) for p in data]

    def get_allusertracks(self):
        data = self.api.listAllUserTracks()
        if not data:
            return []
        return [_parse_track(p) for p in data]

    def get_album_tracks(self, albid):
        data = self.api.getAlbumDetails(album_id=albid)
        if not data:
            return {}
        # Apart from the tracks array, sata contains a copy of the album
        # metadata. Parse it like an album.
        album = _parse_album(data)
        return [_parse_track(t, album) for t in data["tracks"]]

    def get_media_url(self, trackid):
        data = self.api.getTrackById(track_id=trackid)
        if not data or "url" not in data or data["url"] is None:
            uplog("get_media_url: no url for [%s]" % trackid)
            return None
        return data["url"]

    def get_editormoods(self):
        data = self.api.getAvailableMoods()
        return [_parse_album(a) for a in data]

    def get_editorgenres(self):
        data = self.api.getAvailableGenres()
        return [_parse_album(a) for a in data]

    def get_editorthemes(self):
        data = self.api.getAvailableThemes()
        return [_parse_album(a) for a in data]

    def get_alleditorplaylists(self):
        data = self.api.getAllEditorPlaylists()
        return [_parse_album(a) for a in data]

    def get_editor_playlist(self, id):
        data = self.api.getEditorPlaylist(id=id)
        if not data:
            return []
        album = _parse_album(data)
        return [_parse_track(t, album) for t in data["tracks"]]

    def get_playlist_tracks(self, plid):
        data = self.api.playlist_get(playlist_id=plid, extra="tracks")
        return [_parse_track(t) for t in data["tracks"]["items"]]

    def search(self, value):
        data = self.api.quickSearch(search=value)
        albums = []
        for key, value in data.items():
            value["id"] = key
            albums.append(value)
        return [_parse_album(a) for a in albums]

    def searchCategory(self, value, prefix):
        data = self.api.searchCategory(category=prefix, search=value)
        return [_parse_album(a) for a in data]


def encode_prefix(p):
    return base64.urlsafe_b64encode(p.encode("utf-8")).decode("utf-8")


def decode_prefix(p):
    return base64.urlsafe_b64decode(p.encode("utf-8")).decode("utf-8")


def _parse_category(data):
    return Category(
        id=encode_prefix(data["prefix"]), name=data["title"], iid=data["id"]
    )


def _parse_genre(data):
    return Genre(id=encode_prefix(data["prefix"]), name=data["title"], iid=data["id"])


def _parse_artist(json_obj):
    artist = Artist(id=json_obj["id"], name=json_obj["title"].encode())
    return artist


def _parse_album(data):
    # uplog(data)
    # uplog("-_"*30)

    kwargs = {"id": data["id"], "name": data["title"]}
    if "artist" in data:
        kwargs["artist"] = data["artist"]
    if "cover" in data:
        if isinstance(data["cover"], dict):
            for key in ("preview", "master", "thumbnail"):
                if key in data["cover"]:
                    kwargs["image"] = "https://" + data["cover"][key]["file_url"]
                    break
        else:
            kwargs["image"] = data["cover"]
    elif "playlistCover" in data:
        kwargs["image"] = data["playlistCover"]

    a = Album(**kwargs)
    return a


def _parse_track(data, albumarg=None):
    artist = Artist(name=data["artist"])

    kwargs = {
        "id": data["playlistAdd"],
        "name": data["title"],
        "duration": data["playtime"],
        "format": data["format"],
        "artist": artist,
        "available": True,
    }

    if "trackNumber" in data:
        kwargs["track_num"] = data["trackNumber"]
    else:
        kwargs["track_num"] = 0

    if albumarg:
        kwargs["album"] = albumarg
    elif "cover" in data and "album" in data:
        kwargs["album"] = Album(image=data["cover"], name=data["album"])

    # If track has own cover, use it (e.g. for multialbum playlists)
    if "cover" in data:
        kwargs["image"] = data["cover"]

    return Track(**kwargs)


def _parse_playlist(data, artist=None, artists=None):
    kwargs = {
        "id": data["playlist_id"],
        "name": data["playlist_title"],
        "num_tracks": data.get("count_track"),
        "duration": data.get("duration"),
    }
    return Playlist(**kwargs)



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import upmplgmodels
import datetime
import upmpdmeta


def set_track_number(track_number: str, target: dict):
    target['upnp:originalTrackNumber'] = track_number


def set_album_art_from_uri(album_art_uri: str, target: dict):
    target['upnp:albumArtURI'] = album_art_uri


def set_uri(uri: str, target: dict):
    target['uri'] = uri


def set_duration(duration: int, target: dict):
    target['duration'] = str(duration)


def set_disc_number(disc_number: int, target: dict):
    target['discnumber'] = str(disc_number)


def set_channels(channel_count: int, target: dict):
    target['res:channels'] = str(channel_count)


def set_track_title(track_title: str, target: dict):
    target['tt'] = track_title


def set_album_title(album_title: str, target: dict):
    target['upnp:album'] = album_title


def set_description(description: str, target: dict):
    target['dc:description'] = description


def set_album_id(album_id: str, target: dict):
    target['album_id'] = album_id


def set_date(datetime: datetime.datetime, target: dict):
    target['dc:date'] = datetime.strftime("%Y-%m-%d") if datetime else None


def set_class(upnp_class: str, target: dict):
    target['upnp:class'] = upnp_class


def set_artist(artist: str, target: dict):
    target['upnp:artist'] = artist


def set_class_music_track(target: dict):
    target['upnp:class'] = upmplgmodels.Track.upnpclass


def set_class_album(target: dict):
    target['upnp:class'] = upmplgmodels.Album.upnpclass


def set_mime_type(mime_type: str, target: dict):
    target['res:mime'] = mime_type


def set_bit_depth(bit_depth: int, target: dict):
    target['res:bitsPerSample'] = str(bit_depth)


def set_sample_rate(sample_rate: int, target: dict):
    target['res:samplefreq'] = str(sample_rate)


def set_bit_rate(bit_rate: int, target: dict):
    target['kbs'] = str(bit_rate)


def set_object_type_container(target: dict):
    set_object_type('ct', target)


def set_object_type_item(target: dict):
    set_object_type('it', target)


def set_object_type(container_type: str, target: dict):
    target['tp'] = container_type


def set_class_artist(target: dict):
    target['upnp:class'] = upmplgmodels.Artist.upnpclass


def set_raw_metadata(raw_metadata_name: str, metadata_value: str, target: dict):
    if metadata_value is not None and len(metadata_value) > 0:
        # msgproc.log(f"Setting [{raw_metadata_name}]=[{metadata_value}]")
        target[raw_metadata_name] = metadata_value


def set_upmpd_meta(metadata_name: upmpdmeta.UpMpdMeta, metadata_value: any, target: dict):
    v: str = ""
    if metadata_value:
        if isinstance(metadata_value, str):
            v = metadata_value
        elif isinstance(metadata_value, int):
            v = str(metadata_value)
    set_raw_metadata(
        raw_metadata_name=f"upmpd:{metadata_name.value}",
        metadata_value=v,
        target=target)



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cmdtalkplugin
import upmplgutils
import constants
import config
import requests
import os
import secrets
import mimetypes
import glob
import pathlib

import upmpdmeta
import upnp_util

from tidalapi import Quality as TidalQuality
from tidalapi.session import Session as TidalSession
from tidalapi.artist import Artist as TidalArtist
from tidalapi.album import Album as TidalAlbum
from tidalapi.playlist import Playlist as TidalPlaylist
from tidalapi.playlist import UserPlaylist as TidalUserPlaylist
from tidalapi.mix import Mix as TidalMix
from tidalapi.mix import MixV2 as TidalMixV2
from tidalapi.media import MediaMetadataTags as TidalMediaMetadataTags
from tidalapi.media import Track as TidalTrack
from tidalapi.media import Video as TidalVideo
from tidalapi.media import AudioMode as TidalAudioMode
from tidalapi.exceptions import ObjectNotFound

from typing import Callable
from typing import Union

from element_type import ElementType
from element_type import get_element_type_by_name

from item_identifier import ItemIdentifier
from item_identifier_key import ItemIdentifierKey

import identifier_util
import persistence
from played_track import PlayedTrack

from album_sort_criteria import AlbumSortCriteria
from album_adapter import AlbumAdapter
from datetime import datetime

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)

log_unavailable_images_sizes: bool = (upmplgutils.getOptionValue(
    f"{constants.PluginConstant.PLUGIN_NAME.value}log_unavailable_images_sizes",
    "0") == "1")
log_unavailable_image: bool = (upmplgutils.getOptionValue(
                                f"{constants.PluginConstant.PLUGIN_NAME.value}log_unavailable_image",
                                "0") == "1")

default_image_sz_by_type: dict[str, int] = dict()
default_image_sz_by_type[TidalArtist.__name__] = [750, 480, 320, 160]
default_image_sz_by_type[TidalAlbum.__name__] = [1280, 640, 320, 160, 80]
default_image_sz_by_type[TidalPlaylist.__name__] = [1080, 750, 640, 480, 320, 160]
default_image_sz_by_type[TidalMix.__name__] = [1500, 640, 320]
default_image_sz_by_type[TidalMixV2.__name__] = [1500, 640, 320]
default_image_sz_by_type[TidalUserPlaylist.__name__] = default_image_sz_by_type[TidalPlaylist.__name__]


class FavoriteAlbumsMode:

    __element_type: ElementType
    __display_name: str
    __sort_criteria_builder: Callable[[bool], list[AlbumSortCriteria]]
    __descending: bool

    @classmethod
    def create(
            cls,
            element_type: ElementType,
            display_name: str,
            sort_criteria_builder: Callable[[bool], list[AlbumSortCriteria]],
            descending: bool = False):
        obj: FavoriteAlbumsMode = FavoriteAlbumsMode()
        obj.__element_type = element_type
        obj.__display_name = display_name
        obj.__sort_criteria_builder = sort_criteria_builder
        obj.__descending = descending
        return obj

    @property
    def element_type(self) -> ElementType:
        return self.__element_type

    @property
    def display_name(self) -> str:
        return self.__display_name

    @property
    def sort_criteria_builder(self) -> Callable[[bool], list[AlbumSortCriteria]]:
        return self.__sort_criteria_builder

    @property
    def descending(self) -> bool:
        return self.__descending


def __get_image_dimension_list(obj: any) -> list[int]:
    key = type(obj).__name__
    return default_image_sz_by_type[key] if key in default_image_sz_by_type else list()


def get_name_or_title(obj: any) -> str:
    if hasattr(obj, "name"):
        return obj.name
    if hasattr(obj, "title"):
        return obj.title
    return None


def get_cached_image_subdir_list(image_type: str) -> list[str]:
    return [
        constants.PluginConstant.PLUGIN_NAME.value,
        constants.PluginConstant.CACHED_IMAGES_DIRECTORY.value,
        image_type]


def cached_images_exist(album_id: str) -> list[str]:
    document_root_dir: str = config.getWebServerDocumentRoot()
    if document_root_dir:
        sub_dir_list: list[str] = get_cached_image_subdir_list(image_type=TidalAlbum.__name__)
        image_dir: str = ensure_directory(base_dir=document_root_dir, sub_dir_list=sub_dir_list)
        cached_file_name_no_ext: str = f"{str(album_id)}"
        cached_files: list[str] = glob.glob(f"{os.path.join(image_dir, cached_file_name_no_ext)}.*")
        return cached_files if cached_files else []
    return []


def if_cached_exists(album_id: str, image_url: str) -> str:
    return image_url if album_id and len(cached_images_exist(album_id=album_id)) > 0 else None


def get_album_art_url_by_album_id(
        album_id: str,
        tidal_session: TidalSession,
        album: TidalAlbum = None) -> str:
    use_cached: bool = config.get_enable_image_caching()
    if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
        msgproc.log(f"get_album_art_url_by_album_id caching [{use_cached}] "
                    f"for album_id [{album_id}] "
                    f"title [{album.name if album else None}] "
                    f"by [{album.artist.name if album and album.artist else None}] "
                    f"album [{'set' if album else 'not set'}]")
    # use cached?
    cached_files: list[str] = cached_images_exist(album_id=album_id) if use_cached else []
    use_cached = len(cached_files) > 0 if use_cached else False
    if use_cached:
        # try cached!
        # pick newest file
        cached_file: str = __select_newest_file(cached_files)
        cached_file_name_ext: str = os.path.splitext(cached_file)[1]
        # touch the file.
        # msgproc.log(f"About to touch file [{cached_file}] ...")
        pathlib.Path(cached_file).touch()
        # use cached file
        path: list[str] = list()
        sub_dir_list: list[str] = get_cached_image_subdir_list(image_type=TidalAlbum.__name__)
        path.extend(sub_dir_list)
        path.append(f"{str(album_id)}{cached_file_name_ext}")
        cached_image_url: str = compose_docroot_url(os.path.join(*path))
        if config.get_dump_image_caching():
            msgproc.log(f"get_album_art_url_by_album_id [{album_id}] -> [{cached_image_url}]")
        return cached_image_url
    else:
        if config.get_dump_image_caching():
            msgproc.log(f"get_album_art_url_by_album_id [{album_id}] -> cache miss")
    # if we are are, we fallback to normal
    if config.get_dump_image_caching():
        msgproc.log(f"get_album_art_url_by_album_id [{album_id}] -> loading from upstream service is required")
    # album: TidalAlbum = album if album else try_get_album(album_id=album_id, tidal_session=tidal_session)
    if not album:
        album = try_get_album(album_id=album_id, tidal_session=tidal_session)
    return get_image_url(obj=album, refresh=True) if album else None


def get_web_document_root_file_url(dir_list: list[str], file_name: str) -> str:
    if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
        msgproc.log(f"get_web_document_root_file_url for [{dir_list}] [{file_name}]")
    path: list[str] = list()
    path.extend(dir_list)
    path.append(file_name)
    file_url: str = compose_docroot_url(os.path.join(*path))
    if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
        msgproc.log(f"get_web_document_root_file_url for [{dir_list}] [{file_name}] "
                    f"-> [{file_url}]")
    return file_url


def get_image_url(obj: any, refresh: bool = False) -> str:
    if obj is None:
        return None
    if isinstance(obj, TidalTrack):
        # use album instead
        track: TidalTrack = obj
        obj = track.album
    if not config.get_enable_image_caching():
        return __get_image_url(obj)
    document_root_dir: str = config.getWebServerDocumentRoot()
    # webserverdocumentroot is required
    if not document_root_dir:
        return __get_image_url(obj)
    if type(obj) not in [TidalAlbum, TidalArtist, TidalPlaylist, TidalMix]:
        return __get_image_url(obj)
    sub_dir_list: list[str] = [
        constants.PluginConstant.PLUGIN_NAME.value,
        constants.PluginConstant.CACHED_IMAGES_DIRECTORY.value,
        type(obj).__name__]
    image_dir: str = ensure_directory(document_root_dir, sub_dir_list)
    cached_file_names: list[str] = __get_cached_file_names(image_dir, str(obj.id))
    # msgproc.log(f"get_image_url for [{str(obj.id)}] -> [{cached_file_names}]")
    # pick newest file
    cached_file_name: str = (__select_newest_file(cached_file_names)
                             if cached_file_names and len(cached_file_names) > 0
                             else None)
    # touch the file.
    if cached_file_name:
        # msgproc.log(f"About to touch file [{cached_file_name}] ...")
        pathlib.Path(cached_file_name).touch()
    # cached_file_name_ext will include the ".", so it will likely be ".jpg"
    cached_file_name_ext: str = os.path.splitext(cached_file_name)[1] if cached_file_name else None
    # msgproc.log(f"For itemid [{str(obj.id)}] found file [{cached_file_name}]")
    if refresh or not cached_file_name:
        image_url: str = __get_image_url(obj=obj)
        # msgproc.log(f"get_image_url saving to [{image_dir}] [{dest_file}]")
        response = requests.get(image_url)
        content_type = response.headers.get('content-type')
        file_types: list[str] = mimetypes.guess_all_extensions(content_type)
        if file_types and len(file_types) > 0:
            # file_types include the "."
            cached_file: str = os.path.join(image_dir, f"{str(obj.id)}{file_types[0].lower()}")
            if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
                msgproc.log(f"get_image_url got mimetype for [{image_url}] -> "
                            f"[{file_types}] -> "
                            f"saving to [{cached_file}]")
            img_data: bytes = requests.get(image_url).content
            with open(cached_file, 'wb') as handler:
                handler.write(img_data)
            cached_image_url: str = get_web_document_root_file_url(
                dir_list=sub_dir_list,
                file_name=f"{str(obj.id)}{file_types[0].lower()}")
            return cached_image_url
        else:
            msgproc.log(f"Cannot understand file type for item id [{str(obj.id)}], cannot cache.")
            return None
    elif cached_file_name:
        return get_web_document_root_file_url(
                dir_list=sub_dir_list,
                file_name=f"{str(obj.id)}{cached_file_name_ext}")


def __get_cached_file_names(cache_dir: str, item_id: str) -> list[str]:
    cached_file_path: str = os.path.join(cache_dir, item_id)
    item_id_ext: str = os.path.splitext(cached_file_path)[1]
    if item_id_ext:
        # cached_file_path has extension, we convert that to lower case
        item_id_ext = item_id_ext.lower()
        exists = os.path.exists(cached_file_path)
        if exists:
            return [cached_file_path]
    else:
        # no extension in item_id, look what's stored, if any
        matching_files: list[str] = glob.glob(f"{cached_file_path}.*")
        if matching_files and len(matching_files) > 0:
            return matching_files
    return []


def __select_newest_file(file_list: list[str]) -> str:
    if not file_list or len(file_list) == 0:
        return None
    if len(file_list) == 1:
        return file_list[0]
    by_modification_time: list[tuple[str, float]] = []
    curr_file: str
    for curr_file in file_list:
        curr_path = os.path.normcase(curr_file)
        mtime: float = os.path.getmtime(curr_path)
        by_modification_time.append((curr_file, mtime))
    # sort by mtime descending
    by_modification_time.sort(key=lambda x: x[1], reverse=True)
    # get first.
    return by_modification_time[0][0]


def __get_image_url(obj: any) -> str:
    if obj is None:
        return None
    dimension_list: list[int] = __get_image_dimension_list(obj)
    if not dimension_list or len(dimension_list) == 0:
        msgproc.log(f"Type [{type(obj).__name__}] does not have an image sizes list!")
        return None
    current: int
    for current in dimension_list if dimension_list else list():
        try:
            return obj.image(dimensions=current)
        except ValueError:
            if log_unavailable_images_sizes:
                msgproc.log(f"Cannot find image for type [{type(obj).__name__}] "
                            f"id [{obj.id}] Name [{get_name_or_title(obj)}] with size [{current}]")
        except AttributeError as ae_exc:
            msgproc.log(f"Cannot find image for type [{type(obj).__name__}] "
                        f"id [{obj.id}] Name [{get_name_or_title(obj)}] Exception [{ae_exc}]")
    if log_unavailable_image:
        msgproc.log(f"Cannot find image for type [{type(obj).__name__}] id [{obj.id}] "
                    f"Name [{get_name_or_title(obj)}] (any size)")
    return None


def is_mp3(q: TidalQuality) -> bool:
    return q == TidalQuality.low_320k or q == TidalQuality.low_96k


def get_mime_type(track_quality: TidalQuality) -> str:
    if is_mp3(track_quality):
        return "audio/mp3"
    else:
        return "audio/flac"


def is_multidisc_album(album: TidalAlbum) -> bool:
    return album.num_volumes and album.num_volumes > 1


def try_get_track(tidal_session: TidalSession, track_id: str) -> tuple[TidalTrack, Exception]:
    track: TidalTrack = None
    try:
        track = tidal_session.track(track_id)
    except Exception as ex:
        msgproc.log(f"try_get_track failed for track_id [{track_id}] [{type(ex)}] [{ex}]")
        return None, ex
    return track, None


def try_get_album(tidal_session: TidalSession, album_id: str) -> TidalAlbum:
    album: TidalAlbum = None
    try:
        if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
            msgproc.log(f"try_get_album loading album_id [{album_id}] ...")
        album = tidal_session.album(album_id)
        return album
    except ObjectNotFound as onfEx:
        msgproc.log(f"try_get_album could not find album_id [{album_id}] [{type(onfEx)}] [{onfEx}]")
        # TODO more actions? like e.g. remove from favorites
    except Exception as ex:
        msgproc.log(f"try_get_album failed for album_id [{album_id}] [{type(ex)}] [{ex}]")


def try_get_artist(tidal_session: TidalSession, artist_id: str) -> TidalArtist:
    artist: TidalArtist = None
    try:
        artist = tidal_session.artist(artist_id)
    except Exception as ex:
        msgproc.log(f"try_get_artist failed for artist_id [{artist_id}] [{type(ex)}] [{ex}]")
    return artist


class CachedTidalQuality:

    def __init__(
            self,
            bit_depth: int,
            sample_rate: int,
            audio_quality: TidalQuality,
            audio_mode: str = None):
        self._bit_depth: int = bit_depth
        self._sample_rate: int = sample_rate
        self._audio_quality: TidalQuality = audio_quality
        self._audio_mode: str = audio_mode

    @property
    def bit_depth(self) -> int:
        return self._bit_depth

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @property
    def audio_quality(self) -> TidalQuality:
        return self._audio_quality

    @property
    def audio_mode(self) -> str:
        return self._audio_mode


__readable_sr: dict[int, str] = {
    44100: "44",
    48000: "48",
    88200: "88",
    96000: "96",
    176400: "176",
    192000: "192"
}


def __readable_sample_rate(sample_rate: int) -> str:
    if sample_rate in __readable_sr:
        return __readable_sr[sample_rate]
    return None


def get_playlist_by_name(
        tidal_session: TidalSession,
        playlist_name: str,
        create_if_missing: bool = False) -> TidalUserPlaylist:
    tidal_session.user.playlists()
    user_playlists: list[TidalUserPlaylist] = tidal_session.user.playlists()
    # look for the first playlist with the configured name
    # if it does not exist, we create it, if we are allowed to
    listen_queue: TidalUserPlaylist = None
    current: TidalUserPlaylist
    for current in user_playlists if user_playlists else list():
        # msgproc.log(f"get_playlist_by_name processing [{current.name}]")
        if current.name == config.listen_queue_playlist_name:
            listen_queue = current
            break
    if not listen_queue and create_if_missing:
        # we create the playlist
        listen_queue = tidal_session.user.create_playlist(playlist_name)
    return listen_queue


def is_album_in_playlist(
        tidal_session: TidalSession,
        album_id: str,
        playlist_name: str) -> bool:
    listen_queue: TidalUserPlaylist = get_playlist_by_name(
        tidal_session=tidal_session,
        playlist_name=playlist_name)
    if not listen_queue:
        msgproc.log(f"is_album_in_listen_queue for album_id [{album_id}] -> "
                    f"playlist [{config.listen_queue_playlist_name}] not found")
        return False
    # if we find a track from that album, we return True
    offset: int = 0
    limit: int = 100
    while True:
        item_list: list = listen_queue.items(offset=offset, limit=limit)
        msgproc.log(f"Getting [{len(item_list) if item_list else 0}] from playlist "
                    f"[{config.listen_queue_playlist_name}] from offset [{offset}]")
        if not item_list or len(item_list) == 0:
            # we are finished
            # msgproc.log(f"No more items from offset [{offset}], we are finished")
            break
        count: int = 0
        for current in item_list if item_list else list():
            count += 1
            # must be a track
            if not isinstance(current, TidalTrack):
                continue
            track: TidalTrack = current
            if track.album.id == album_id:
                # msgproc.log(f"Found track [{track.id}] from album [{album_id}], returning True")
                return True
        if count < limit:
            # msgproc.log(f"Found [{count}] instead of [{limit}] from offset [{offset}], we are finished")
            break
        offset += limit
    return False


def album_playlist_action(
        tidal_session: TidalSession,
        album_id: str,
        playlist_name: str,
        action: str) -> TidalAlbum:
    listen_queue: TidalUserPlaylist = get_playlist_by_name(
        tidal_session=tidal_session,
        playlist_name=playlist_name,
        create_if_missing=True)
    # remove anyway, add to the end if action is add
    remove_list: list[str] = list()
    offset: int = 0
    limit: int = 100
    while True:
        item_list: list = listen_queue.items(offset=offset, limit=limit)
        # msgproc.log(f"Getting [{len(item_list) if item_list else 0}] from playlist "
        #             f"[{config.listen_queue_playlist_name}] from offset [{offset}]")
        if not item_list or len(item_list) == 0:
            # we are finished
            msgproc.log(f"No more items from offset [{offset}], we are finished")
            break
        count: int = 0
        for current in item_list:
            count += 1
            # must be a track
            if not isinstance(current, TidalTrack):
                continue
            track: TidalTrack = current
            if track.album.id == album_id:
                # msgproc.log(f"Found track [{track.id}] from album [{album_id}], returning True")
                remove_list.append(track.id)
        if count < limit:
            # msgproc.log(f"Found [{count}] instead of [{limit}] from offset [{offset}], we are finished")
            break
        offset += limit
    for track_id in remove_list:
        listen_queue.remove_by_id(media_id=track_id)
    album: TidalAlbum = tidal_session.album(album_id=album_id)
    # add if needed
    if constants.ListeningQueueAction.ADD.value == action:
        # add the album tracks
        media_id_list: list[str] = list()
        t: TidalTrack
        for t in album.tracks():
            media_id_list.append(t.id)
        listen_queue.add(media_id_list)
    return album


def is_tidal_album_stereo(album: TidalAlbum) -> bool:
    # missing -> assume STEREO
    media_metadata_tags: list[str] = album.media_metadata_tags
    return is_stereo(media_metadata_tags)


def __audio_modes_has_stereo(audio_modes: list[str]) -> bool:
    if not audio_modes or len(audio_modes) == 0:
        return False
    current: str
    for current in audio_modes:
        if TidalAudioMode.stereo == current:
            return True
    return False


def is_stereo(media_metadata_tags: list[str]) -> bool:
    # nothing available -> assume STEREO
    if not media_metadata_tags or len(media_metadata_tags) == 0:
        return True
    return not __only_of(
        media_metadata_tags=media_metadata_tags,
        hit=[TidalMediaMetadataTags.dolby_atmos])


def __only_of(media_metadata_tags: list[str], hit: list[str]) -> bool:
    if not media_metadata_tags or len(media_metadata_tags) == 0:
        return False
    current: str
    for current in media_metadata_tags:
        if current not in hit:
            return False
    return True


def not_stereo_skipmessage(album: TidalAlbum) -> str:
    return (f"Skipping album with id [{album.id}] [{album.name}] by [{album.artist.name}] "
            f"because: [{album.media_metadata_tags}]")


def __get_best_quality(media_metadata_tags: list[str]) -> str:
    if not media_metadata_tags or len(media_metadata_tags) == 0:
        return None
    if TidalMediaMetadataTags.hi_res_lossless in media_metadata_tags:
        return TidalQuality.hi_res_lossless
    if TidalMediaMetadataTags.lossless in media_metadata_tags:
        return TidalQuality.high_lossless
    if TidalMediaMetadataTags.dolby_atmos in media_metadata_tags:
        return TidalQuality.low_96k
    return None


def try_get_all_favorites(tidal_session: TidalSession) -> list[TidalAlbum]:
    favorite_list: list[TidalAlbum] = list()
    offset: int = 0
    limit: int = 100
    while True:
        some: list[TidalAlbum] = None
        try:
            some: list[TidalAlbum] = tidal_session.user.favorites.albums(limit=limit, offset=offset)
        except Exception as ex:
            msg: str = f"Cannot get favorite albums from offset [{offset}] [{type(ex)}] [{ex}]"
            raise Exception(msg)
        some_len: int = len(some) if some else 0
        if some_len > 0:
            favorite_list.extend(some)
        if some_len < limit:
            break
        # another slice maybe
        offset += limit
    return favorite_list


def get_quality_badge(
        album: TidalAlbum,
        cached_tidal_quality: CachedTidalQuality) -> str:
    return get_quality_badge_raw(
        audio_modes=album.audio_modes,
        media_metadata_tags=album.media_metadata_tags,
        audio_quality=album.audio_quality,
        cached_tidal_quality=cached_tidal_quality)


def get_quality_badge_raw(
        audio_modes: list[str],
        media_metadata_tags: list[str],
        audio_quality: str,
        cached_tidal_quality: CachedTidalQuality) -> str:
    # msgproc.log(f"get_quality_badge type(audio_modes) -> {type(audio_modes) if audio_modes else 'none'}")
    # TODO maybe map DOLBY_ATMOS to say Atmos
    if audio_modes and not __audio_modes_has_stereo(audio_modes):
        return ",".join(audio_modes)
    tidal_quality: TidalQuality = __get_best_quality(media_metadata_tags)
    if not tidal_quality:
        tidal_quality = audio_quality
    stream_info_available: bool = (
        cached_tidal_quality and
        cached_tidal_quality.bit_depth and
        cached_tidal_quality.sample_rate)
    bit_depth: int = cached_tidal_quality.bit_depth if cached_tidal_quality else None
    sample_rate: int = cached_tidal_quality.sample_rate if cached_tidal_quality else None
    ext_badge: str = (f"{bit_depth}/{__readable_sample_rate(sample_rate)}"
                      if stream_info_available else None)
    badge: str = None
    if TidalQuality.hi_res_lossless == tidal_quality:
        if ext_badge:
            badge = ext_badge
        else:
            badge = f"HD {ext_badge}" if ext_badge else "MAX"
    elif TidalQuality.high_lossless == tidal_quality:
        if not stream_info_available:
            badge = "Lossless"
        else:
            if bit_depth == 16 and sample_rate == 44100:
                badge = "CD"
            elif bit_depth == 16 and sample_rate == 48000:
                badge = "16/48"
            else:
                badge = ext_badge
    elif TidalQuality.low_320k == tidal_quality:
        badge = "Low/320"
    elif TidalQuality.low_96k == tidal_quality:
        badge = "Low/96"
    if config.display_quality_badge:
        msgproc.log(f"get_quality_badge q:[{tidal_quality}] "
                    f"b:[{bit_depth}] "
                    f"s:[{sample_rate}] -> [{badge}]")
    return badge


def track_only(obj: any) -> any:
    if obj and isinstance(obj, TidalTrack):
        return obj


def get_mix_or_playlist_items(
        tidal_session: TidalSession,
        tidal_obj_id: str,
        obj_type: str,
        limit: int,
        offset: int) -> list[any]:
    underlying_type: ElementType = get_element_type_by_name(element_name=obj_type)
    if underlying_type == ElementType.PLAYLIST:
        playlist: TidalPlaylist = tidal_session.playlist(tidal_obj_id)
        return playlist.tracks(limit=limit, offset=offset)
    elif underlying_type == ElementType.MIX:
        mix: TidalMix = tidal_session.mix(tidal_obj_id)
        items: list[any] = mix.items()
        items = items if items else list()
        sz: int = len(items)
        if offset >= sz:
            return list()
        if offset + limit > sz:
            # reduce limit
            limit = sz - offset
        return items[offset:offset + limit]
    else:
        raise Exception(f"Invalid type [{obj_type}]")


def load_unique_ids_from_mix_or_playlist(
        tidal_session: TidalSession,
        tidal_obj_id: str,
        tidal_obj_type: str,
        id_extractor: Callable[[any], str],
        max_id_list_length: int,
        previous_page_last_found_id: str = None,
        item_filter: Callable[[any], any] = lambda x: track_only(x),
        initial_offset: int = 0,
        max_slice_size: int = 100) -> tuple[list[str], list[any], int, bool, str]:
    if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
        msgproc.log(f"load_unique_ids_from_mix_or_playlist mix_or_playlist_id [{tidal_obj_id}] "
                    f"tidal_obj_type [{tidal_obj_type}] "
                    f"initial_offset [{initial_offset}] max_slice_size [{max_slice_size}]")
    last_offset: int = initial_offset
    id_list: list[str] = list()
    obj_list: list[any] = list()
    load_count: int = 0
    skip_count: int = 0
    finished: bool = False
    last_found: str = None
    while len(id_list) < max_id_list_length:
        max_loadable: int = max_slice_size
        if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
            msgproc.log(f"load_unique_ids_from_mix_or_playlist mix_or_pl [{tidal_obj_id}] "
                        f"tidal_obj_type [{tidal_obj_type}] "
                        f"loading [{max_loadable}] from offset [{last_offset}] "
                        f"load_count [{load_count}] skip_count [{skip_count}] ...")
        item_list: list[any] = get_mix_or_playlist_items(
            tidal_session=tidal_session,
            tidal_obj_id=tidal_obj_id,
            obj_type=tidal_obj_type,
            limit=max_slice_size,
            offset=last_offset)
        if not item_list or len(item_list) == 0:
            # no items, we are finished
            finished = True
            break
        slice_count: int = 0
        for i in item_list:
            slice_count += 1
            last_offset += 1
            item = item_filter(i) if item_filter is not None else i
            if item:
                id_value: str = id_extractor(item)
                # already collected?
                if (id_value and
                    (not previous_page_last_found_id or id_value != previous_page_last_found_id) and
                        (id_value not in id_list)):
                    id_list.append(id_value)
                    obj_list.append(item)
                    last_found = id_value
                if len(id_list) == max_id_list_length:
                    # we are finished
                    break
            else:
                skip_count += 1
        load_count += len(item_list)
        # did we extract less than requested?
        if len(item_list) < max_slice_size:
            # we are finished in this case
            finished = True
            break
    return id_list, obj_list, last_offset, finished, last_found


def get_webserver_cached_images_path() -> list[str]:
    return [
        constants.PluginConstant.PLUGIN_NAME.value,
        constants.PluginConstant.CACHED_IMAGES_DIRECTORY.value]


def get_webserver_static_images_path() -> list[str]:
    return [
        constants.PluginConstant.PLUGIN_NAME.value,
        constants.PluginConstant.STATIC_IMAGES_DIRECTORY.value]


def get_plugin_code_subpath() -> list[str]:
    return ["cdplugins", constants.PluginConstant.PLUGIN_NAME.value]


def get_plugin_code_static_images_subpath() -> list[str]:
    return get_plugin_code_subpath() + [constants.PluginConstant.PLUGIN_IMAGES_DIRECTORY.value]


def get_plugin_static_images_abs_path() -> str:
    plugin_data_dir: str = upmplgutils.getOptionValue("pkgdatadir")
    plugin_path: list[str] = get_plugin_code_static_images_subpath()
    p: str
    for p in plugin_path:
        plugin_data_dir = os.path.join(plugin_data_dir, p)
    return plugin_data_dir


def ensure_directory(base_dir: str, sub_dir_list: list[str]) -> str:
    curr_sub_dir: str
    curr_dir: str = base_dir
    for curr_sub_dir in sub_dir_list:
        new_dir: str = os.path.join(curr_dir, curr_sub_dir)
        # msgproc.log(f"checking dir [{new_dir}] ...")
        if not os.path.exists(new_dir):
            msgproc.log(f"creating dir [{new_dir}] ...")
            os.mkdir(new_dir)
        # else:
        #     msgproc.log(f"dir [{new_dir}] already exists.")
        curr_dir = new_dir
    return curr_dir


def get_docroot_base_url() -> str:
    host_port: str = (os.environ[constants.EnvironmentVariableName.UPMPD_UPNPHOSTPORT.value]
                      if constants.EnvironmentVariableName.UPMPD_UPNPHOSTPORT.value in os.environ
                      else None)
    doc_root: str = (os.environ["UPMPD_UPNPDOCROOT"]
                     if "UPMPD_UPNPDOCROOT" in os.environ
                     else None)
    if not host_port or not doc_root:
        return None
    return f"http://{host_port}"


def compose_docroot_url(right: str) -> str:
    doc_root_base_url: str = get_docroot_base_url()
    return f"{doc_root_base_url}/{right}" if doc_root_base_url else None


def get_oauth2_credentials_file_name() -> str:
    return os.path.join(upmplgutils.getcachedir(constants.PluginConstant.PLUGIN_NAME.value), constants.oauth2_credentials_file_name)


def oauth2_credential_file_exists() -> bool:
    return os.path.exists(get_oauth2_credentials_file_name())


def get_pkce_credentials_file_name() -> str:
    return os.path.join(upmplgutils.getcachedir(constants.PluginConstant.PLUGIN_NAME.value), constants.pkce_credentials_file_name)


def pkce_credential_file_exists() -> bool:
    return os.path.exists(get_pkce_credentials_file_name())


def is_instance_of_any(obj: any, type_list: list[type]) -> bool:
    t: type
    for t in type_list if type_list else list():
        if isinstance(obj, t):
            return True
    return False


def get_all_mix_or_playlist_tracks(mix_or_playlist: Union[TidalPlaylist, TidalMix]) -> list[TidalTrack]:
    return (get_all_playlist_tracks(playlist=mix_or_playlist)
            if isinstance(mix_or_playlist, TidalPlaylist)
            else get_all_mix_tracks(mix=mix_or_playlist))


def get_all_playlist_tracks(playlist: TidalPlaylist, max_tracks: int = None) -> list[TidalTrack]:
    result: list[TidalTrack] = []
    offset: int = 0
    default_limit: int = 100
    limit: int = min(default_limit, max_tracks) if max_tracks else default_limit
    finished: bool = False
    while not finished:
        slice: list[TidalTrack] = playlist.items(offset=offset, limit=limit)
        slice_len: int = len(slice) if slice else 0
        offset += slice_len
        finished = slice_len < limit
        item: Union[TidalTrack, TidalVideo]
        for item in slice if slice else []:
            if isinstance(item, TidalTrack):
                result.append(item)
                if max_tracks and len(result) >= max_tracks:
                    break
    return result


def get_albums_from_tracks(track_list: list[TidalTrack]) -> list[TidalAlbum]:
    return list(map(lambda t: t.album, track_list))


def get_artists_from_tracks(track_list: list[TidalTrack]) -> list[TidalArtist]:
    return list(map(lambda t: t.artist, track_list))


def get_all_mix_tracks(mix: TidalMix) -> list[TidalTrack]:
    result: list[TidalTrack] = []
    items: list[TidalTrack] = mix.items()
    item: Union[TidalTrack, TidalVideo]
    for item in items:
        if isinstance(item, TidalTrack):
            result.append(item)
    return result


def create_mix_or_playlist_all_tracks_entry(
        objid: any,
        element_type: ElementType,
        thing_id: str,
        thing: Union[TidalPlaylist, TidalMix],
        all_tracks: list[TidalTrack]) -> dict[str, any]:
    all_tracks_identifier: ItemIdentifier = ItemIdentifier(
        ElementType.ALL_TRACKS_IN_PLAYLIST_OR_MIX.getName(),
        thing_id)
    all_tracks_identifier.set(key=ItemIdentifierKey.UNDERLYING_TYPE, value=element_type.getName())
    all_tracks_id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(all_tracks_identifier))
    all_tracks_entry = upmplgutils.direntry(all_tracks_id, objid, "All Tracks")
    select_item: Union[TidalTrack | TidalVideo] = secrets.choice(all_tracks) if all_tracks else None
    msgproc.log(f"create_mix_or_playlist_all_tracks_entry select [{select_item is not None}] "
                f"type [{type(select_item) if select_item else None}]")
    upnp_util.set_album_art_from_uri(
        album_art_uri=get_image_url(select_item),
        target=all_tracks_entry)
    return all_tracks_entry


class PageLinkIdentifier:

    def __init__(self, value: str, api_path: str, category_title: str):
        self.__value: str = value
        self.__api_path: str = api_path
        self.__category_title: str = category_title

    @property
    def value(self) -> str:
        return self.__value

    @property
    def api_path(self) -> str:
        return self.__api_path

    @property
    def category_title(self) -> str:
        return self.__category_title


def get_cached_audio_quality(album_id: str) -> CachedTidalQuality:
    played_track_list: list[PlayedTrack] = persistence.get_played_album_entries(album_id)
    if not played_track_list or len(played_track_list) == 0:
        return None
    # get first
    played_track: PlayedTrack = played_track_list[0]
    # audio_mode: str = played_track.audio_mode
    audio_quality: TidalQuality = played_track.audio_quality
    # audio quality not available? fix when possible
    if not audio_quality:
        # identify hi_res_lossless
        if ((played_track.bit_depth and played_track.bit_depth > 16) and
           (played_track.sample_rate and played_track.sample_rate > 48000)):
            return CachedTidalQuality(
                bit_depth=played_track.bit_depth,
                sample_rate=played_track.sample_rate,
                audio_quality=TidalQuality.hi_res_lossless)
        # identify hi_res
        if played_track.bit_depth and played_track.bit_depth > 16:
            # just hires
            return CachedTidalQuality(
                bit_depth=played_track.bit_depth,
                sample_rate=played_track.sample_rate,
                audio_quality=TidalQuality.hi_res_lossless)
    # catch invalid combinations
    bit_depth: int = played_track.bit_depth
    sample_rate: int = played_track.sample_rate
    if bit_depth == 16 and sample_rate in [44100, 48000]:
        if audio_quality in [TidalQuality.hi_res_lossless]:
            # invalid!
            # reset audio_quality to None to avoid false hires identification
            audio_quality = None
    return CachedTidalQuality(
        bit_depth=bit_depth,
        sample_rate=sample_rate,
        audio_quality=audio_quality)


def get_album_disc_and_track_counters(album: AlbumAdapter) -> str:
    disc_count: int = album.num_volumes
    result: str = (f"{disc_count} Disc{'s' if disc_count > 1 else ''}, "
                   if disc_count and disc_count != -1 else "")
    result += (f"{album.num_tracks} Track{'s' if album.num_tracks > 1 else ''}"
               if album.num_tracks and album.num_tracks != -1 else "")
    return result


def date_to_display(dt: datetime) -> str:
    return f"{dt.year:04d}/{dt.month:02d}/{dt.day:02d}"


def add_album_adapter_metadata(album_adapter: AlbumAdapter, target: dict[str, any]):
    # upv
    universal_product_number: int = album_adapter.universal_product_number
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.UNIVERSAL_PRODUCT_NUMBER, str(universal_product_number), target)
    # artist id
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ARTIST_ID, album_adapter.artist_id, target)
    # album id
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_ID, album_adapter.id, target)
    # album title
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_TITLE, album_adapter.name, target)
    # version
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_VERSION, album_adapter.version, target)
    # explicit
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_EXPLICIT_STATUS, 'yes' if album_adapter.explicit else 'no', target)
    # copyright
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.COPYRIGHT, album_adapter.copyright, target)
    # audio quality
    cached_tidal_quality: CachedTidalQuality = get_cached_audio_quality(
        album_id=album_adapter.id)
    badge: str = get_quality_badge_raw(
        audio_modes=album_adapter.audio_modes,
        media_metadata_tags=album_adapter.media_metadata_tags,
        audio_quality=album_adapter.audio_quality,
        cached_tidal_quality=cached_tidal_quality)
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_QUALITY, badge, target)
    # artist names
    artist_names: list[str] = []
    artist_list: list[TidalArtist] = album_adapter.artist_list if album_adapter.artist_list else []
    curr_artist: TidalArtist
    for curr_artist in artist_list:
        artist_names.append(curr_artist.name)
    # only show if this is meaningful
    if len(artist_list) > 1:
        artist_name_list: str = ", ".join(artist_names)
        upnp_util.set_upmpd_meta(
            upmpdmeta.UpMpdMeta.ALBUM_ARTIST,
            artist_name_list,
            target)
        if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
            msgproc.log(f"Artists for [{album_adapter.id}] set to [{artist_name_list}] ...")
    # duration
    duration_sec: int = album_adapter.duration
    if album_adapter.duration and album_adapter.duration != -1:
        upnp_util.set_upmpd_meta(
            upmpdmeta.UpMpdMeta.ALBUM_DURATION,
            upmpdmeta.get_duration_display_from_sec(duration_sec),
            target)
    # disc and tracks
    disc_track_counters: str = get_album_disc_and_track_counters(album_adapter)
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_DISC_AND_TRACK_COUNTERS, disc_track_counters, target)
    # available release date
    available_release_date_str: str = date_to_display(album_adapter.available_release_date)
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_AVAILABLE_RELEASE_DATE, available_release_date_str, target)
